# -*- coding: utf-8 -*-
# ipadic と jumandic は品詞の定義等々においてかなり異なるため、
# ipadic と jumandic をダイレクトに共存させることは出来ない。
# また、 jumandic には役に立たなかったりかえって投入によって
# 解析に支障をきたすエントリが大量に含まれている。ただ ipadic
# にはないエントリも数多いことも事実なので、ここから使えるもの
# だけ抽出して、「ユーザ辞書として」使いたい。
#
# 「スクリプト」の体裁はしているけれど、実際のところは「抽出
# 作業記録」そのものである。ゆえに「あなたのために汎用に」
# 整理するつもりはあんましない。そもそも jumandic の最終バージョン
# は 2013 年のものであって、2019年10月の今からみてもう 7年
# 近くまるまる更新がなく、おそらく今後も更新されないと思われる。
# だから基本的に「いったん抽出してしまったらこのスクリプトは
# 用済み」でいいわけだ。かろうじて「抽出方法をちょっと変える」
# とか、あなたが「オレ式に抽出」のためのベースくらいにはなる
# であろう。そもそもがここでやっている抽出方法は、あくまでも
# 私の考えに基くものであって、あなたのものではない。気に入らな
# ければ自分流の抽出をすると良かろう。(特に「接尾」の扱いは
# 相当入り組んでる。あなたと同じ考えかどうか正直わからない。
# 多分かなり違うだろう。)
#
# なお、以下は抽出を試みない:
# * Assert
#   そもそも欲しいものがない。ここあるのは ipadic のコンセプト
#   による解析でも特段不満はないのでは? たとえば「であったろう」
#   は、たぶん ipadic 流儀では「で あっ た ろう」と分解するはず。
#   「あっ」は無論「ある」の連用タ接続。普通はこれでいいわけだ。
# * Auto
#   これは配布用辞書エントリではないんだろうなぁという言うナゾの
#   ファイル。
#     桃井はるこだ,1381,1381,4360,形容詞,*,ナ形容詞,基本形,桃井はるこだ,桃井はるこだ,自動獲得:テキスト
#   こんなんをいちいち「形容詞」として登録する意味はないが、
#   かといって「固有名詞をここから抽出したれ」と思っても今度は
#     ファンタジックだったろう,1339,1339,9495,形容詞,*,ナ形容詞,ダ列タ系推量形,ファンタジックだ,ファンタジックだったろう,自動獲得:テキスト
#   と、固有名詞とは限らないのが区別出来ない状態で入ってる。
#   ゆえに、基本的に機械的に分類しようとするのはかなり無理がある。
#   出来るのは語幹の抽出のみ。それではまったく価値がない。
# * Wikipedia
#   Auto と同じくで、さらに輪をかけてヒドいともいえるし、あるいは
#   逆に、少しは Auto よりは区別可能で抽出できるかも、と言える。
#   たとえば「旧跡」として分類されているところに落語家の名前が
#   大量に含まれている、などがあるが、丹念にみれば何かめぼしいもの
#   が抽出出来る可能性はある。16万レコードの中から。
#   つまりは「何もここから抽出せんでも良いのでは」ということ。
# * 助詞
#   まず、ipadic の left-id.def (right-id.def) を参照してもらうと
#   わかるが、ipadic では助詞は「そこにあるもので全て」である。
#   ユーザ辞書に自由に追加出来る想定に、ほぼなっていない。出来
#   ないわけではないが、文脈ID の目的外使用が必要である。結局
#   語ひとつひとつに対して文脈IDを与えているので、jumandic から
#   持ってこようとするなら「left-id.def にあるものから同じものを
#   選ぶ」かもしくは「left-id.def にあるものから近いものを選ぶ」
#   ということになるが、前者はもちろん無意味、後者は「労多くして
#   功少なし」の典型。機械判定はほとんど不可能なので、「この
#   スクリプトっぺーなにかそれらしきもの」で実現するにしても、
#   ばりっばりのハードコーディングでしか出来ない。そしてそんだけ
#   頑張った挙句に、めぼしいものはほとんど抽出出来ない。
# * 助動詞
#   助詞と話は似ているけれど、こっちはもっとシンプルに、jumandic
#   にある助動詞はとても数が少なく、目視確認で全部「別になくてもいい
#   かな」と思うものばかりである。マッピングは助詞以上に大変で
#   ある。「代表表記」で言えばたった9個(utf-8で扱えないもの含めると
#   もうあと3個くらいだったかしら)しかないものに対してかける
#   労力ではない。(「大変」は、想像よりも遥かに「大変」。)
# * 判定詞
#   たったひとつのエントリ「だ」だけ(活用するので15)。「判定詞」
#   が ipadic のどの品詞に対応するのかを調べてもいないが、
#   ipadic に同じ役割のものはあるはずだ。ないとは言わせない。
#   まぁいらんでしょ。(当たり前だが労力はかなりのもの。たった
#   一つのために。)
# * カ変動詞、カ変動詞来、ipadic での「サ変・スル」
#   自立・非自立ともに、全部 ipadic に既存。因果応報的変種もなし。
#
# 下の方でもう一度、「このスクリプトだとうったえているなんか
# それっぺーもの」の構造や考え方の説明の中で触れるのだが、
# jumandic エントリにはびっくりするほど「いらない子」が多い。
# 下での説明では「因果応報」のヘンチクリンな変種で説明している
# jumandic の基本コンセプトは「綴り違いに対して寛容」である。
# ただの寛容ではない。「いやいや、そりゃぁやり過ぎだ」という
# 水準である。綴り違いというよりは「綴り間違い」に属するような
# ものも結構あるし、漢字が含まれるものについては必ずオールひら
# がなバージョンが入っている。おそらく規則に基いてバリエーション
# を機械生成したんだと思うのだが、そんな綴り、するわきゃなかろ、
# てのが大量に入ってる。
#
# そうした「いらない子」たちが必ず問題を起こすかといえば別に
# そんなことはないし、その変種たちの、少なくとも名詞以外の
# もののうちの漢字を含むものの半分以上は役に立つものたちである。
# たとえば「申し付ける」の変種として「申つける」は普通いらんだ
# ろう。こういう「間違い」は、やりたいことによっては「未知語」
# として切り出されるほうが適切な場合もある。けれども「盛り立てる」
# に対する「盛立てる」は、こういうのは割と人によってばらつく
# ので役に立つかもしれない。一般的にいって、送り仮名が違うと
# 読み手が混乱しうるものや現実にほかの意味と衝突する語は、変種
# は生じにくい。たとえば「細い」を「細かい」のつもりで書く人が
# ほぼ皆無なのは、無論「細い(ほそい)」があるからである。jumandic
# の生成する変種は、このことに関して意識的なのかそうでないのか
# 判断がつかない。まさに「細い(こまかい)」はいない一方で、
# 「卸しね」(卸値)はしっかりいる。これを「卸値」のことだと
# 一発で認識出来る人は半分以下だろう。(おそらく「御死ね」だと
# 思うだろう。)
#
# なので、この「なんかスクリプトとか言っちゃってるよこの子」は、
# こうしたものを「せめて区別可能なように」頑張っている。けれども
# それにも限度があって、原則として「ipadic に既存のものに対する
# 変種かどうか」しか役に立つ分類は出来ない。完全に新規エントリ
# 扱いになるものどうしの変種はあんまし区別出来ない。これは「代表
# 表記」として代表にされているものが、決して「最頻」が選ばれて
# いるわけではないことが一番の原因である。具体例としてはたとえば
# 「かけがえのない」。「掛け替えの無い」が代表表記とされているが、
# 日常的にはこれをわざわざ漢字で書く人の方が稀だろう。
#
# そんなわけで、この「スクリプトだってばよ」が生成する辞書エントリ
# の使い方、使い道の考え方について。直前で言ったように、jumandic
# のエントリにはなんなら「超絶に迷惑な」エントリが大量に含まれて
# いる。「たかがひらがな、されどひらがな」。地名のひらがなは無害
# そうに見えるけれど、たとえば「ほっかいどう」が問題を起こすことは
# ほぼ皆無だろうが、「ちば」は違う。「いちばん」は jumandic エントリ
# がいないうちはおおむね「いちばん」として切り出すだろうが、「ちば」
# エントリによって、混同可能性が生じる。これを多少は分類するわけだ
# が、この「みせてもらおうか、連邦軍のスクリプトとやらを」による
# 分類は、おそらく多くの人にとって不十分だ。ぜってーいらねーという
# ヤツが役に立つ分類に入り、役に立つ欲しいエントリが役に立たない
# 分類に入る(一応元のコストを参考に振り分けてる)。この生成結果を
# 全部そのまま受け容れるのも別に悪いことではないのだが、もっと良い
# のはきっと「あれ? こんなのも入ってなかったんだ」に気付いた語が
# あった際にこの「あれれスクリプトとかいってら」が吐き出した結果
# から探して、それだけを使う、てことじゃないだろうか。少なくとも
# 元の jumandic ソースから使うよりは、すべて ipadic 流儀に変換して
# あるこちらが遥かに楽だろう。(オリジナルは特に「タ形/タ系」が邪魔
# だ。)
#
# 今はやりの人工知能ではないのだから、こうした伝統的な辞書ってのは、
# 「量が多ければ多いほどいい」というもんではなくて、適用対象に
# 適した辞書を使うのが一番良い結果をもたらす。2ch/5ch のような
# 壊れた日本語ばかりを対象にしたいなら、こいつで生成する全部が
# 役に立つだろうし、そうでないなら「選択的に取り込む」のがいい
# んじゃないだろうか。
#
import os
import posixpath
import io
import sys
import re
import csv
import unicodedata
import subprocess
import itertools
from collections import defaultdict
from glob import glob
import logging


_khmap = {
    chr(i + ord("ぁ")): chr(i + ord("ァ"))
    for i in range((ord("ん") - ord("ぁ")) + 1)}
def _h2k(s):
    return "".join(
        [_khmap.get(c, c) for c in s])
_hkmap = {
    chr(i + ord("ァ")): chr(i + ord("ぁ"))
    for i in range((ord("ン") - ord("ァ")) + 1)}
def _k2h(s):
    return "".join(
        [_hkmap.get(c, c) for c in s])


try:
    __mecab_config_exec = 'bash -c "`type -p mecab-config` {}"'
    _mecab_config = {}
    for q in ("--exec-prefix", "--dicdir",):
        _mecab_config[q] = subprocess.check_output(
            __mecab_config_exec.format(q),
            shell=True).strip().decode()
except subprocess.CalledProcessError as e:
    _mecab_config = {
        "--exec-prefix": "c:/Program Files (x86)/MeCab",
        "--dicdir": "c:/Program Files (x86)/MeCab/dic"
        }
_dict_index_bin = os.path.join(
    _mecab_config["--exec-prefix"], "bin", "mecab-dict-index")
_sysdict_dir = os.path.join(_mecab_config["--dicdir"], "ipadic")

if ":" in _mecab_config["--exec-prefix"]:  # Windows
    _mecab_config["--sysdic-src-encoding"] = "cp932"
else:
    _mecab_config["--sysdic-src-encoding"] = "utf-8"


# ==================================================================
# スクリプト全体の説明
# ==================================================================
#
# スクリプトがやること自体は至って「単純」なのだけれど、「簡単」かと
# 言うとさほどでもない。jumandic の辞書ソースは「内容はかなり未整理、
# 構造はやや未整理」なのが原因。内容の未整理度に関しては、おそらく
# 実験的なプロジェクトだからこれでいいのだと思う。受け容れて自分で
# なんとかするしかない。ただ「構造の未整理」がちょっと処理上問題で、
# この未整理が原因で、妙な重複エントリをたくさん生み出すハメになる。
#
# そういうわけで、この「スクリプトのようなナニか」は多段構えである。
#
# 第一段階は「シンプルなプリプロセス」。
#
# ここで行うのは「(jumandic 定義の)品詞だけによるファイルの再構成」
# と「原形と代表表記によるグルーピング」、読みの ipadic 流儀への
# 変換(最小限)、そして「取捨選択第一段階」。(あくまでも「原則として」。
# 実際はこれ以上のこともやってる。以後の説明も同様。)
#
# 最初のは、まずは複数のファイルに同じエントリが散在していることが
# あり、また逆に本来分離されているべきものが一つのファイルに集まっ
# ているので、これを再構成するということ。(ただしこの時点では「品詞」
# は jumandic 定義のものなので、これでも ipadic 方式に変換後には
# ひとつの「品詞」が複数品詞に分岐することがある。これは後続の処理
# がやるのでここでは気にしない。)
#
# ただこの件が「jumandic の罪」かどうかは正直わからなくて。辞書
# ソース形式の csv 全てが「配布辞書用エントリ」かどうかがわからない
# のだけれど、シード用辞書と配布用辞書が全部含まれてるなら、まぁ
# こういうことは起こってしまうわけで。ともかく「わかんないけど
# 扱えそうなものは全部扱う」と考えればこうなる、と。
#
# 二つ目は、 jumandic のそのノリである。基本的に「無難・穏健」とは
# ほど遠くて、「綴り違いバージョンやひらがなバージョンに対応しちゃる
# →一括で全バリエーション機械生成だっ」という、まぁ結構実験的と
# いうかなんというか。システム辞書向けではないノリ。ひらがなバージョン
# は例えば「北海道」なら「ほっかいどう」も同時に存在している。
# 綴り違いは、「リアルワールドでもよくある実用的なエントリ」も
# かなりある一方で、機械生成であるがゆえに「因果応報」「因果応ほう」
# 「因果おう報」「いん果応報」以下ry. と延々「普通はそう綴らない」
# というバージョンが入っている。
#
# これの何が問題かというと、ひとつには「フォーマルな文書に近づく
# ほどこうしたエントリが無用な混同をもたらす」、つまり「迷惑だ」
# ということ。もうひとつは私のこの「スクリプトのようなナニモノか」
# のコンセプトと相性が悪いこと。何かというと、「ipadic に既存の
# ものはいらない」てこと。つまりこの「妙ちくりんなバリエーション」
# は ipadic には存在しないわけだ。ゆえに、なんなら「ヘンなバージョン
# だけを選択的に取り込む」ハメになる。そういうことがしたいんではない。
#
# で、jumandic と ipadic は定義がかなり違っているので、その読み替えを
# しなければならない、のは当然(これは第二段階以降)。ただ、この
# 「読み替え」する前から既に「これはいらない」とわかるものがある。
#
# 「いらない」の理由は複数あるが、根本的なところでは「ipadic 流儀
# の定義に基く解析をする限りにおいては必要ない」か、もしくは
# そもそも対応する left-id, right-id にマッピングできないもの。
# 例えば品詞細分類で「タ系」とある活用は、ipadic 流儀では完全に
# 無用の長物。仮に欲しいとしても、 ipadic に取り込みたいなら
# left-id.def 等々の定義から変更を要するのであって、それはもはや
# ipadic ではなくなる。また、「これは jumandic からはいらない」
# というもので、「人名,姓」。私はこれについては「myoujijiten」
# さえあればいいので、jumandicから抽出する意味がない。
#
# 第一段階については、プリプロセス処理結果を一時ファイルとして
# 書き出す。第二段階の処理はこれを入力とする。あとこれはしょうがない
# のだけれど、utf-8 として(Python が)扱えないエントリが少量あって、
# これはこの時点で捨てる。
#
# 第二段階は「品詞等の読み替え」。
#
# ここでは絞込みのことは考えずに、ひたすら対応をマッピングしていく。
# この結果も一時ファイルとして書き出す。第三段階はこれを入力とする。
#
# 第三段階は ipadic とのマッチングによる「読み替え・取捨選択第二段階」。
#
# 原則として ipadic に既存のエントリは抽出しない、というのが一つ
# だが、その「既存」の扱いは結構複雑である。単純に「品詞～活用形3」
# の合致だけを合致とみなすと、相当数が新規エントリ扱いになる。
# けれどもたとえば、最もわかりやすいものだと、「洞爺湖」は ipadic
# では「固有名詞,一般」として登録されている。対して jumandic では
# 「地名」扱いである。これはどちらが正しいということではなく、
# カテゴライズが違っている、というだけなのだが、こういう「分類
# し直せば既存」というものについて、新規エントリ扱いにするか、
# 重複とみなすかはものによって違う。同じく名詞を例にすれば、
# 「副詞可能 vs サ変接続」などは、おそらく新規扱いにして良い。
#
# 第三段階の複雑さのもう一つの一因は無論「因果おう報」の扱いである。
# マッチングは、第一段階で行ったグルーピング単位で行うことになる。
# つまり、このグループ単位で「どれか一つでも合致扱いのものがいれば
# それはグループとして合致である」とする必要がある。また、そうして
# 合致とみなされたものは、「全てをリジェクトする」のではなく、
# バリエーションの方は新規扱いとするのだが、その一つ一つについて
# 対応する ipadic と同じ定義に読み替える必要がある。(「合致」
# は「完全一致」ではない。細分類レベルが少々違うものは合致と
# みなす。)
#
# 最後の第四段階は、最終的に (my_mecab_userdict_build.py 前提の)
# 辞書ソースを生成すること、だが、ここでも最終分類を行った上で
# 書き出す。
# 分類は、一つには「ipadic にあるものの変種かどうか」ということ。
# さらには「代表表記と異なるかどうか」。「ipadic にあるものの変種
# かどうか」で吐き出すディレクトリ自体を分離してしまうが、
# 「代表表記と異なるかどうか」は ipadic に既存かどうかによらず
# 各エントリに、 my_mecab_userdict_build.py のコメント形式で
# 識別出来るようにする。もう一つは「ひらがなのみのエントリかどうか」。
# 適用対象のテキストのフォーマルっぷり、カジュアルっぷりによって、
# 「ipadic にあるものの変種、ひらがなバージョン」のありがたみと迷惑
# さは激変する。崩れた表現が多い対象、たとえば 5ch の議論をターゲット
# にしたいならおそらくこの変種は役に立つだろう。逆に「普通のテキスト」
# 相手の場合は、不愉快な混同による悪影響の方が強く出て、ありがたみの
# 方が薄れる可能性が高い。利用時は、実際に解析させてみて利用するか
# どうかを決めるといいだろう。
#

# ==================================================================
# 第一段階関連のもろもろ
# ==================================================================
_rgx_dh_extra = re.compile(r"代表表記:([^/\s]+)[^\s]*")
def _firststep_filter(jdent):
    # 最初の取捨選択。
    #
    # 「ソフトウェアの正しいあり方」として本当はここで処理するのは
    # NG であろう、というものを多数含んでいる。ただそれってのは
    # 「将来性を考えるならば」の話。けど、「mecab-jumandic-7.0-20130310」
    # から7年ほったらかしなわけだから、そもそもこいつに将来なんてない
    # と考えたほうがいい。
    #

    # うま,1192,1192,9681,形容詞,*,イ形容詞アウオ段,語幹,うまい,うま,連語
    sur = jdent[0]  # 表層形
    hin = jdent[4]  # 品詞
    hd1 = jdent[5]  # 品詞細分類1
    kk1 = jdent[6]  # 品詞細分類2、ipadict で言うところの「活用形1」。
    kk2 = jdent[7]  # 品詞細分類3、ipadict で言うところの「活用形2」。
    of = jdent.pop(8)  # 原形
    extra = jdent[-1].strip()

    # 代表表記は、書かれているものに関しては、活用する語の
    # 場合、「基本形の原形」。活用しない語の場合、変種で
    # なければ「表層形 == 原形 == 代表表記」となるが、
    # 活用語の場合は「表層形 != 原形 == 代表表記」である。
    # すなわち、変種かどうかは表層形ではなく原形で見るべし。
    dh = of
    if "代表表記" in extra:
        dh = _rgx_dh_extra.search(extra).group(1).strip()
        extra = _rgx_dh_extra.sub("", extra).strip()
    elif "連語" in extra or (hin, hd1) == ("名詞", "数詞",):
        # 代表表記が書かれていないので特に問題なのは、Rengo.csv のもの。ここに
        # しかないものがないなら悩まなかったんだけれど、残念ながら少し
        # めぼしいのがいる。問題は代表表記がないばっかりに、ひらがなエントリ
        # が完全新規もの扱いになってしまうことだったりする(「ほんとう」など)。
        # つまりここでやりたいのは「欲しいものが欲しい」ためじゃなくてむしろ
        # 「欲しくないかもしれないものを区別したい」ため、なわけよ。不毛
        # だよな。
        # これのものはここでマニュアルで(必要最小限)補う。注意すべきなのは、
        # Rengo.csv 以外にある重複エントリ(のいずれか)と一致させないといけ
        # ないこと。私の意見ではないよ、なんかどマイナーなのを代表表記に
        # しとるなー、と思ったとしても。
        # (必要最小限、というのは、今この処理を書いている時点において既に
        # 後続の処理がおおむね完成していて、その処理前提にすればいい、
        # てこと。だから実際の処理順と「スクリプトらしきそれっぺーやつ」
        # を書いた順番は全然トップダウンじゃない。よーするにここ、最後の
        # 最後に書いてる、ワタシは。)
        if hd1 in ("普通名詞", "形式名詞", "組織名", "サ変名詞",
                   "判定詞",) or hin in ("特殊",):
            pass
        elif hin != "動詞" and hd1 != "数詞":
            # ↑おバカな処理でこなす都合、分けて処理しないとやり過ぎる
            for rx, repl in (
                # 「旨い」「美味い」など色々あるが、Rengo.csv では「上手い」。
                # ほかのも同じ。
                (re.compile(r"(うまい|上手い).*$"), "上手い"),
                (re.compile(r"(よい|良い).*$"), "良い"),
                (re.compile(r"(ない|無い).*$"), "無い"),
                (re.compile(r"(ながい|長い).*$"), "長い"),
                (re.compile(r"(ほんとう|本当).*$"), "本当だ"),
                (re.compile(r"(おおい|多い).*$"), "多い"),
                (re.compile(r"(すくない|少ない).*$"), "少ない"),

                #
                (re.compile(r"(あまり|余り).*$"), "余り"),
                (re.compile(r"(あがる|上がる).*$"), "上がる"),
                #(re.compile(r"(いま|今)$"), "今"),
                #(re.compile(r"(しかけ|仕掛け|シカケ)$"), "仕掛け"),

                (re.compile(r"(すこし|少し)$"), "少し"),
                (re.compile(r"(よく|良く)$"), "良く"),

                (re.compile(r"(よう|様)$"), "様"),
                (re.compile(r"(り|人)$"), "人"),
                (re.compile(r"(どおり|通り)$"), "通り"),
                (re.compile(r"(尽くし|づくし)$"), "尽くし"),

                (re.compile(r"(おる|居る)$"), "居る"),

                #(re.compile(r"(また|又)$"), "又"),

                (re.compile(r"(ため|為)$"), "為"),
                (re.compile(r"(お|御)$"), "御"),
                (re.compile(r"(ほど|程)$"), "程"),
                (re.compile(r"(ごとし|如し)$"), "如し"),

                #(re.compile(r"^(.*)(ように|様に)$"), r"\1ように"),
                ):
                if rx.match(of):
                    dh = rx.sub(repl, of)
                    break
        elif hd1 == "数詞":
            for rx, repl in (
                (re.compile(r"^(すう|数)$"), "数"),
                (re.compile(r"^(なん|何)$"), "何"),
                (re.compile(r"^(いく|幾)$"), "幾"),
                ):
                if rx.match(of):
                    dh = rx.sub(repl, of)
                    break
        else:  # 動詞
            for rx, repl in (
                (re.compile(r"(ある|有る).*$"), "有る"),
                (re.compile(r"(あがる|上がる).*$"), "上がる"),
                (re.compile(r"(いる|居る).*$"), "居る"),
                (re.compile(r"(いう|言う).*$"), "言う"),
                (re.compile(r"(よる|因る).*$"), "因る"),
                (re.compile(r"(つづける|続ける).*$"), "続ける"),
                (re.compile(r"(かかわる|係わる).*$"), "係わる"),
                (re.compile(r"(なる|鳴る).*$"), "鳴る"),
                (re.compile(r"(おく|置く).*$"), "おく"),
                (re.compile(r"(つく|付く).*$"), "付く"),
                (re.compile(r"(とる|取る).*$"), "取る"),
                (re.compile(r"(やむ|止む).*$"), "止む"),
                (re.compile(r"(える|得る).*$"), "得る"),
                ):
                if rx.match(of):
                    dh = rx.sub(repl, of)
                    break
    elif "人名" not in extra and hin not in ("助動詞", "助詞"):
        # ほかにも書かれていないものはあるのだが、たとえば「指示詞」では
        # ひらがな表記に代表表記がなく、漢字の方が「代表表記」でひらがな
        # に歩み寄っていて問題ない。(たとえば「その様に」の代表表記が
        # 「そのように」になっている。)あと幸いというかなんというか、
        # 人名は「太郎/太ろう/た郎/たろう」なんてバカげたことはしてない
        # ので代表表記そのものが書かれてない。そりゃそーだ。「太ろう」
        # は「ふとろう」だろ常考。
        if re.match(r"^[ぁ-んー]+$", sur) and len(of) > 1:  # and kk2 == "語幹":
            logging.debug("%r, %r", of, jdent)
            pass

    # 記号。
    if hd1 == "記号":
        # jumandic にいるものは ipadic と違い「アルファベット」区分を
        # していない。そして「アルファベット」に区分されるべきエントリは
        # 全て ipadic にいる。全部アスキーなので。(ipadic ではギリシャ
        # 文字なども「アルファベット」として入れている。)
        # ゆえに、「「アルファベット」に区分されるべきエントリ」を予め
        # 捨てる。(ipadic との重複を取り除いた上で最終的に残るのは、
        # 少量の記号と大量の「顔文字」。)
        sur_n = unicodedata.normalize("NFKC", sur)
        if re.match(r"^[A-Za-z]$", sur_n):
            return None

    # 「人名,姓」は、myoujijiten さえあればいいのでいらない。
    # 外国人名はどうかといえば、「人名:外国」として姓・名
    # 区別されてない。ので「外国人の姓だけは欲しい」と
    # 思ってもそうはいかない。
    if re.search(r"人名:[^:]+:姓", extra):
        return None

    if hin == "接尾辞" and sur in ("区", "郡", "町"):
        # なぜか jumandic では「Noun.koyuu.csv」(固有名詞) の中にこの3
        # つが紛れ込んでいる。いずれも ipadic の方に既存だが、「区」だけは
        # 細分類が違う。 jumandic の「Noun.koyuu.csv」はこれら3つを全て
        # 「接尾辞,名詞性特殊接尾辞」で登録しようとする。ipadic では、
        # 「郡」と「町」を「名詞,接尾,地域」として、「区」を「名詞,接尾,
        # 助数詞」と「名詞,接尾,地域」の二つを登録しようとする。つまり
        # 「地域」カテゴリの接尾辞としては ipadic には全て入っているので
        # あって、完全に無視してよい。
        return None

    #
    if kk1 == "*" and \
            re.match(r"^[ァ-ンヴー・]+$", sur) and \
            not re.match(r"^[ァ-ンヴー・]+$", dh) and \
            sur == re.sub(r"[^ァ-ンヴー・]", "", dh):
        # この「汎用的」な判定でつかまるのは実は「地名:国」のみである。
        # そしてそれこそが捨てたいもの。「略語も入ってりゃ便利っしょ」
        # ルールに基づいて、こんなの:
        #   代表表記: アラブ首長国連邦
        #   表層形  : アラブ
        #   分類    : 地名:国
        # が9ついる。いや…「アラブ」は国じゃなかろう。ドミニカ、コンゴ
        # なんぞはまぁ「国」扱いでもいいんだが、アラブは絶対違うし、
        # マーシャル、ミクロネシアは微妙、バチカンもちとあやうい。
        # つーわけで、そもそもこの「略語も入れときゃうんたら」には
        # そもそも共感出来てないわけなんだけれど、そもそもね、カタカナ
        # 語てのは、周縁にきちんとひらがな・漢字があれば、ちゃんと
        # かたまりとして「未知語」で切り出されるわけよ。分類としては
        # 「固有名詞: 組織」かもしれんけれど、少なくとも「分かち書き」
        # したいだけの MeCab 利用にとっては、「コンゴ」が国として入って
        # ないことが悪影響になることはまずない。あとちなみに、この
        # 代表表記でのくくり方そのものが、「このスクリプトとか言いやがっ
        # て」の処理上ちょいと厄介なのだ。そんなのもあってね、「ワタシ
        # はこんなんいらん」てことで。
        #
        #logging.error((yomi, sur, dh, of, hd1, extra))
        return None

    # ------------------------------------------------------------------
    #
    # 活用に関する扱い関係でいらないものを捨てていく。話は結構繊細。
    #

    # 良い、無い、恰好良い|格好良い|…の3つについては、当初は扱う
    # つもりだった。この3つとも ipadic ではかなりスペシャルな
    # 扱いをされていて、扱おうとすると非常に面倒なうえに、
    # 「だっているじゃん」てことで、いらんでしょ? いらんよね?
    # いらないとは言わせない。(「かっこいい」は ipadic で「不変化
    # 型活用」という特殊なタイプになってる。)
    kakkoii = (
        "かっこいい", "カッコいい", "かっこよい",
        "格好いい", "恰好良い", "格好良い",
        )
    if (dh in ("無い", "良い")) or (dh in kakkoii) or (of in kakkoii):
        return None

    # 「タ形/タ系」は ipadic 流儀ではそもそも必要ない。ので、
    # 捨ててしまえばいいといえばいいのだが、(自家発電方式で
    # はなく)「左から右へ」アプローチを維持しようと思う場合、
    # ちょっとだけ代用品として残しておくと楽。その「本当は
    # いらない無用の長物」のうち、一部を除いて全て捨てる。
    if re.search(r"タ系", kk2) and kk2 not in (
        "タ系推量形", "タ系省略推量形"):
        return None

    if (hd1, kk1) == ("形容詞性名詞接尾辞", "イ形容詞アウオ段"):
        # 該当は「-くさい」「-っぽい」「-い」の3つだけ。「-くさい」
        # 「-っぽい」は ipadic に「形容詞,接尾」として既存。「-い」
        # は考えてもいいかもしれないけれど、私のこの「スクリプト…」
        # の処理が工夫しないと間違っちゃうようなわずらわしいやつ。
        # あんましこれ一つのためだけに頑張りたくない。
        return None

    # 活用語について、 jumandic のは「活用語??」というものがいくつか
    # ある。ipadic 流儀で考える場合、活用させなくても良いやつである。
    # そして、それ含め全ての活用語で「語幹」のエントリがある。語幹
    # 相当のエントリは ipadic にはない。ので、「用途がない限りは」原則
    # 捨てれば良いのだが、逆に「活用語、なのかいな、こやつ」に関しては
    # そのまさに「語幹さえあればいい」。
    # 具体的には、jumandic での活用には「気味」に対して「じゃ」「やろう」
    # など独特なものがあるけれど、それらも ipadic 流儀なら「いらない子」
    # である上に、「気味」は ipadic 方式だとシンプルに(活用しない)
    # 「接尾,一般」としていて、ワタシもそれでいいと思うわけだ。
    # 「用途がない限りは」に関してだが、形容詞でのマッピングでの代用
    # 目的で使うので、(まずは)「形容詞のうち ipadicでも活用するやつ」
    # に関する語幹は維持する。
    gokandakegahosiihito = (
        kk1 in ("ナノ形容詞", "ナ形容詞", "ナ形容詞特殊") )
    if not gokandakegahosiihito and kk2 == "語幹":
        if hin != "形容詞":
            return None
    elif gokandakegahosiihito and kk2 != "語幹":
        return None

    if hd1 == "動詞性接尾辞":
        # 動詞性接尾辞-サ変動詞は「する」だけなので、いらない。
        # 動詞性接尾辞-動詞性接尾辞うる型は「うる(/うれば)」だけなので、
        # いらない。
        # 動詞性接尾辞-カ変動詞来はほんとに「来」だけなのでいらない。
        # (ipadic には「動詞,接尾,副詞可能」として登録されている。)
        if kk1 in ("サ変動詞", "動詞性接尾辞うる型", "カ変動詞来"):
            return None

    if kk1 == "動詞性接尾辞ます型":
        # 「動詞性接尾辞ます型」として「存じ」「御座い」とそのひらがな
        # 版の4つ。うち、後者は ipadic では「御座る(助動詞,五段・ラ行
        # 特殊)」と「ます(助動詞,特殊・マス)」でほとんど事足りるので、
        # いらないと思う。「存じ」に関しては、ipadic では「ぞんずる
        # (動詞,自立,サ変・−ズル)」がいるのだが、これが「ぞんじ」に
        # 活用しない。(「-ズル」なのだから。)なので jumandic のが価値
        # がある…、かというと、そうでもない。ipadic 流儀で考える場合
        # はむしろ、必要なのは「ぞんじ」「存じ」「存知」という語 (名詞
        # の副詞可能、助動詞語幹とか、あるいは形容動詞?)なのではないか、
        # と考えても良さそうである。jumandic 流儀でやろうとするとそもそ
        # も「存じ上げます」「存じ奉り上げます」などの変種もあれこれ
        # 登録しなければならなくなる。もしくは、「存じ(動詞,非自立,
        # 一段)」という手はある。
        # ちなみに「存知である」とか「存知している」はあまり日常で
        # 使う言葉ではないが、「STEINS;GATE」(ゲーム/アニメ)の鈴羽と
        # いうキャラクターが「ぞんちしている」と言う言い方をしてた。
        # (現代語に疎く未来の軍隊系の堅苦しい言葉遣いをする、という
        # キャラ設定。)「存知」が存知であれば、こういうのにも対応出来
        # るだろう。いずれにせよ、ともに「あえてjumandic から抽出する
        # 意味はない」ので、これは無視する。
        # (この種の言葉遣いに対応することそのものは非常に意義深いこと
        # だけれども、なにもここからじゃなくてもいい。)
        return None

    if kk1 == "子音動詞カ行促音便形" or dh == "行く":
        # 「行く」「逝く」「いく」のみであり、全部 ipadic に既存なので
        # いらない。(ややこしいが「子音動詞カ行促音便形」でなくて
        # 代表表記「行く」にひっつく「ゆく」がいる。こいつも一緒くたに
        # いらない。というかそっちだけ扱おうとすると面倒なことになる。
        return None

    if kk1 != "*" and re.match(r"^[ぁ-んー]$", of):
        # 接尾のやつが問題なのだが、接尾であろうがなかろうが、
        # どっちにしてもちょいとトラブルメーカなエントリに
        # なりうる上に、わたしのこの「スクリプトと申し上げて
        # いるかのようなナニモノか」の処理でもちぃと面倒な
        # ことが起こるもんで。
        return None

    # 品詞云々をみるまでもなく、個別でいらないものたち。
    # 基本的にここいらは「さじかげん」というやつで、
    # 「「先-駈ける」でいいから「先駆ける」はいらんだろう」
    # みたいになるかならないかは、結構ものによる人による。
    # 唯一の正解はない。(たぶん今の例での「先」単独で混乱
    # しやすいかどうか、くっついていることによって読みが
    # 変わるかどうか、あたりは基準のひとつだろうとは思う。
    # たとえば「-付ける」は個人的には「-付ける」という
    # 非自立動詞がいて、「痛め付ける」は「痛め(他動詞とか)」
    # + 「付ける(非自立)」みたいになればいいんじゃないかな
    # と思うんだけれど、ipadic でも「-付ける」は各々独立
    # して自立動詞として登録してある。)
    if re.match(r".+(できる|出来る)$", dh):
        # 該当するのは実際には「全うできる」のみ。
        # ipadic の場合、「全うする」というサ変動詞と「全う」という
        # サ変接続名詞がいるので、「全うできる」がいないことによって
        # 悲劇的なことが起こるということはない。「全う(サ変接続)」
        # + 「出来る(動詞)」とかになるだろう。
        # この jumandic 流儀を正義とする場合、「完遂できる」のような
        # ものもチマチマ追加しなければならんだろう。
        # (オールひらがなバージョンだと「まっとう」は形容動詞語幹
        # 名詞しか入ってないが、状況はあんまり違わないと思うよ。)
        return None

    return (dh, of, sur, hin, hd1, kk1, kk2, extra)


# 「形容詞」はほっといてもかなり膨大な量があるわけだが、 ipadic
# では「名詞,ナイ形容詞語幹」というものを導入して、量が増えるの
# をおさえている。これが大変便利なのだが、 jumandic はそういう
# ことをしていない。で、「～ない」形容詞のどれが「名詞,ナイ形容
# 詞語幹」になりうるのか、についての「機械的な判定方法」はないと
# 思う。無論概念的には「-ない」を取り除いた部分が名詞かどうか、
# ということだけれど、そもそも「抽象」を表現するものが名詞なのか
# どうかについては、おそらく人それぞれで意見が違うと思う。さすがに
# 「おっか-ない」の「おっか」を名詞ととらえる人は現代人では少ない
# であろうが、「思いがけ-ない」についてはどうだろうか? 私は「思い
# がけ」も抽象名詞と考えていいんじゃないかと思う。けれども ipadic
# では「思いがけ」は「名詞,ナイ形容詞語幹」にしてない。「しょざい
# ない」などについてもそうである。これについてはもういわゆる
# 「コンセンサス」なんてないと思うから、「私がナイ形容詞語幹だと
# 思うもの」というオレ基準で決めようと思う。
#
# おそらく「「-ない」を取り除いてもそんなにヘンチクリンでない」
# ことと、あとたぶん「-がない」「-もない」「-げない」に交換可能かも
# 基準になりそうな気がする。あと「かなわない」がそうでないのは
# これは原義が「適う(動詞)」の否定だから。これだとわかるものはもちろん
# 対象外になる。
#
# そういうわけで、 _rgx_adj_nai は「名詞,ナイ形容詞語幹にしたい」一覧
# である。処理の都合正規表現だがそれそのものは本質ではない。
# 「にしたいもの」を変更したくなったらここを保守する、すなわち
# 「およそスクリプトを名乗るナニモノか」にあるまじき構造だが、
# そんなことは知ったことではない。これは「私の作業記録」に過ぎない
# のであるから。
#
# なお、これの有無による「実用性の違い」についても人によってかなり
# 違うと思う。例えば「下らない」について『「下ら-ない」に分解する
# など万死に値する』と感じる人は当然いると思う。人間が、分析目的
# 以外でそのように分解することなどなく「下らないは下らない」から
# である。のみならず、要素が小さくて出現頻度が高いものほど混同を
# もたらしやすいというのもある。なんであれ、はっきりいって MeCab
# で何をしたいかによる。それともう一つ言うと、 jumandic の例のアレ
# (「因がおう報」)の面からも、「名詞,ナイ形容詞語幹」にすげかえると
# 辞書のサイズが多少削減出来る、という面もある。(以下の正規表現に
# それが現れている。)
#
# この正規表現は第二段階以降でも使う。
_rgx_adj_nai = re.compile(r"""
(
あどけ
|
おぼつか
|
かたじけ
|
ぎこち
|
さりげ
|
しどけ
|
しょうが
|
すげ
|
だらし
|
とんでも
|
にべ
|
ふがい
|
もったい
|
やるせ
|
(?:仕方|しかた|し方)
|
(?:仕様|しよう|し様|仕よう)
|
(?:何気|なにげ|何げ|なに気)
|
(?:余儀|よぎ|よ儀|余ぎ)
|
(?:危なげ|あぶなげ)
|
(?:口さが|くちさが)
|
(?:呆気|あっけ|あっ気|呆け)
|
(?:味気|あじけ|味け|あじ気)
|
(?:大人気|おとなげ|おとな気|大人げ)
|
(?:如才|じょさい|じょ才|如さい)
|
(?:心許|こころもと|心もと|こころ許)
|
(?:忍び|しのび)
|
(?:思い掛け|おもいがけ|おもい掛け|思いがけ)
|
(?:情け|なさけ)
|
(?:惜しみ|おしみ)
|
(?:所在|しょざい|所ざい|しょ在)
|
(?:拠所|よんどころ)
|
(?:揺るぎ|ゆるぎ)
|
(?:果てし|はてし)
|
(?:極まり|きわまり)
|
(?:満遍|まんべん|まん遍|満べん)
|
(?:物足り|ものたり|物たり|もの足り)
|
(?:申し訳|もうしわけ|申しわけ|もうし訳)
|
(?:素っ気|そっけ|素っけ|そっ気)
|
(?:致し方|いたしかた|致しかた|いたし方)
|
(?:造作|ぞうさ|造さ|ぞう作)
|
(?:違い|ちがい)
|
(?:遣る方|やるかた|やる方|遣るかた)
|
(?:限り|かぎり)
|
(?:面目|めんぼく|めん目|面ぼく)
|
(?:頼り|たより)
|
(?:飽き足り|あきたり|あき足り|飽きたり)
)
([のがも]?[無な]い)""", flags=re.X)


class __yomi_norm(object):
    # _yomi_cmp のほうも参照。完全なものは書けない。ほんとに
    # 最小限のみ。
    #
    def __init__(self, alphabets):
        self._repls = alphabets
        self._repls["＆"] = "アンド"  # ipadicで「ト」と読ませちゃってるので。

    def __call__(self, hin, sur, yomi):
        if hin not in ("記号", "特殊"):
            # 「トリニダード・トバゴ」の読みは「トリニダードトバゴ」
            # である。ipadic では。「・」自身でない限りは、読みからは
            # 「・」が取り除かれている。
            yomi = yomi.replace("・", "")
            # ipadic では、アルファベットのみの語にも読みを全て与えて
            # いる。(「Ｈ」が含まれる語で「エッチ」と「エイチ」の両方
            # 管理するというような謎仕様があるが、これは気にしないこと
            # にする。)
            for fr, to in self._repls.items():
                yomi = yomi.replace(fr, to)
        yomi = _h2k(yomi)

        # 「標準化」の範疇からはズレるんだけれども、地名になぜだか
        # 「ュウシ」祭り。何かの作業時に混入したミス? おそらく「ジュウシ
        # ワカマツ」だののエントリからコピペするなりした際になんらかミス
        # したのではないか。拗音始まりの語なんかないので、これはまぁ
        # 「安全に取り除ける」と考えてよかろう。
        if yomi.startswith("ュウシ"):
            yomi = yomi[len("ュウシ"):]
        # そして「北九州」の読みが「キタキュウ」。これも読みとして幸い
        # 一意なので、安全にリプレイス出来る。「ュウシ」は北九州の断片
        # だろうかねぇ?
        if yomi == "キタキュウ":
            yomi = "キタキュウシュウ"

        return yomi

# この初期化は _main にて。
_yomi_norm = None


def _firststep(args):
    # TODO: argsで「第一段階除外」出来るようにしつつ、除外時以外はまっさらに。
    # (第二段階も同じく。)
    blddir = posixpath.join(args.builddir, "1st")
    bldtmpdir = os.path.join(blddir, "_tmp")
    if not os.path.exists(bldtmpdir):
        os.makedirs(bldtmpdir)
    # "AuxV.csv" にだけ UnicodeError になるエントリが少しだけ。
    # 予めこれは捨てておく。
    auxv_tmpout = os.path.join(bldtmpdir, "AuxV.csv")
    with io.open(auxv_tmpout, "wb") as fo:
        with io.open(os.path.join(args.jumandicdir, "AuxV.csv"), "rb") as fi:
            for line in re.split(rb"\r?\n", fi.read().strip()):
                try:
                    fo.write((line.decode("utf-8").strip() + "\n").encode("utf-8"))
                except UnicodeError as e:
                    pass

    targets = [auxv_tmpout]
    for fn in (
        "ContentW.csv",
        "Demonstrative.csv",
        "Emoticon.csv",
        "Noun.hukusi.csv",
        "Noun.keishiki.csv",
        "Noun.koyuu.csv",
        "Noun.suusi.csv",
        "Postp.csv",
        "Prefix.csv",
        "Rengo.csv",
        "Special.csv",
        "Suffix.csv",
        ):
        targets.append(os.path.join(args.jumandicdir, fn))
    hinshi_mapping = {
        # 細分類のみならず品詞からして ipadic と違っているものがあるので注意。
        '形容詞': "Adj",
        'ナ・ナノ形容詞': "Adj_na_nano",
        'タル形容詞': "Adj_taru",
        '形容詞(ない)': "Adj_nai",
        '形容詞(不変化型)': "Adj_huhenka",
        '動詞(付属動詞候補)': "Verb_suf",
        '動詞(カ変)': "Verb_ka",  # 処理対象外だけれどプリプロセスとしては出すだけ出しとく。
        '動詞(サ変)': "Verb_sa",

        '動詞(子音)子音動詞ラ行イ形': "Verb_si_ra_special",
        '動詞(子音)ワ行文語音便形': "Verb_si_wa_special",
        '動詞(子音)': "Verb_si",

        '動詞(母音)': "Verb_bo",
        '連体詞': "Adnominal",
        '副詞': "Adverb",
        '感動詞': "Interjection",
        '接続詞': "Conjunction",
        '名詞': "Noun",
        '接頭辞': "Prefix",
        '助詞': "Postp",  # 処理対象外だけれどプリプロセスとしては出すだけ出しとく。
        '接尾辞': 'Suffix',
        '指示詞': "_Demonstrative_",  # ipadic での代名詞と連体詞と副詞
        '特殊': "Symbol",
        '助動詞': "Auxil",  # 処理対象外だけれどプリプロセスとしては出すだけ出しとく。

        # 処理対象外な上に、いっとー最初の「一文字フィルタ」で、結果的には出力されんのよ
        '判定詞': "_Hanteishi_",
        }

    all_entries = defaultdict(set)
    _reader = itertools.chain.from_iterable(
        (csv.reader(io.open(fn, encoding="utf-8")) for fn in targets))
    for jdent in _reader:
        res = _firststep_filter(jdent)
        if not res:
            continue
        (dh, of, sur, hin, hd1, kk1, kk2, extra) = res

        keys = [hin]

        # 「ナノ・ナ・タル」の説明は込み入ってるので他の場所で
        # 説明する。
        if kk1 in ("ナノ形容詞", "ナ形容詞", "ナ形容詞特殊",):
            keys = ["ナ・ナノ形容詞"]
        elif kk1 in ("タル形容詞",):
            keys = ["タル形容詞"]

        elif hin == "形容詞" and kk1 == "イ形容詞アウオ段" and _rgx_adj_nai.match(dh):
            # _rgx_adj_nai のコメント参照。ここではほかでは
            # _firststep_filter でやっていることをやってしまう。
            # ここでするのは、extra にそれとわかる補足を補うことと、
            # 基本形以外を捨てることである。(出力ファイルにも
            # 影響させたい関係で、ここの方がふさわしいので。)
            keys = ["形容詞(ない)"]
            extra = " ".join(["(:→名詞/ナイ形容詞語幹:)", extra])
            if kk2 != "基本形":
                continue

        elif hin == "動詞":
            # ipadic での「サ変・スル」になるものはたった一つ
            # 「する」であり、jumandic にも当然一つ、ゆえに
            # 抽出不要。(活用する語の共通事項として「ipadic
            # にない活用は無視。)
            if kk1 == "サ変動詞" and of == "する":
                continue

            #
            upperkeys = (
                # この分割ははっきりいって人間のためだけ。
                # 後続の処理で分割されてないから困るとか
                # はまったくない。
                (re.compile("カ変動詞.*"), "動詞(カ変)"),
                (re.compile("[サザ]変動詞.*"), "動詞(サ変)"),
                (re.compile("子音動詞ラ行イ形"), "動詞(子音)子音動詞ラ行イ形"),
                (re.compile("子音動詞ワ行文語音便形.*"), "動詞(子音)ワ行文語音便形"),
                (re.compile("子音動詞.*"), "動詞(子音)"),
                (re.compile("母音動詞"), "動詞(母音)"),
                )
            #
            for k, v in upperkeys:
                if k.match(kk1):
                    if v == "動詞(母音)":
                        # 「母音動詞」という言葉の意味はよくわか
                        # らんが、ipadic (と普通の考え方)で「一段」
                        # になるやつら。「一段・クレル」と「一段・得る」
                        # が別途あって、前者はそのまま ipadic に既存
                        # なので捨てればいい。ややこしいのは「得る」で、
                        # ipadic では「一段」のと「一段・得る」の二つの
                        # バージョンがある。ただ jumandic では
                        # 通常の「一段」しかないので、結果的には
                        # 「得る」に関しては何も考えなくていい。
                        if re.match(r"^[呉く]れる$", dh):
                            # ipadic の「一段・クレル」になるやつ。
                            keys = []
                        else:
                            keys = [v]
                    else:
                        keys = [v]
                    break

            if "付属動詞" in extra:
                # 「候補」と言ってる通りで、どうも作業用メモのような
                # もので、辞書の定義として何かしら対応することをしてる
                # わけではない、と思う。ただこれが書かれてるやつは
                # 「非自立 or 接尾」にしても良いものたち。
                keys.append("動詞(付属動詞候補)")

        # 先頭に「代表表記」追加、「原形」は「代表表記」の
        # 次に移動。直後のソートと後続(第二段階以降)のグルーピング
        # のため。(読みについては、ちょっとだけ ipadic 流儀に馴染ま
        # せておく。(_yomi_norm のコメント参照。)
        newent = [dh, of]
        newent.extend(jdent[:-2])
        newent.append(_yomi_norm(hin, sur, jdent[-2]))
        newent.append(extra)
        for key in keys:
            all_entries[key].add(tuple(newent))

    # 第一段階の最後はあまり本質的ではない「ごみ掃除」を行なう。
    # そもそもが jumandic の辞書エントリが収められている csv
    # ファイルの「どれが正なのかがわからない」ことが根本にはある
    # のだが、同一エントリが複数ファイルにまたがっている。それら
    # が文字列としての完全一致ならば害はないのだが、最後の補足
    # 事項が書かれている列のみが違う状態で収められている場合があ
    # る。「Rengo.csv」内のものはどうやら全てそうらしい。そういう
    # ことであれば最後の列を無視してもいいわけだが、「人名:日本:性」
    # など一部情報を使うために、単純には捨てられない。ゆえに、
    # それら不統一な補足列を一つにまとめてしまう。また、この「スク
    # リプトを主張以下ry...」がおそらく今後も利用することのない
    # 補足については、単純に捨てる。
    def _sortkey(e):
        # ここでのソートは補足列をまとめるのに十分であればいい。
        return (tuple(e[:-1]), -len(e[-1]), e[-1])
    def _cleanup(hin, ents):
        for g in itertools.groupby(
            sorted(list(ents), key=_sortkey),
            lambda x: x[:-1]):  # 補足列以外でグルーピング
            entrycols = list(g[0])  # 補足列以外
            # 補足列を一つにまとめる。
            extra = " ".join(
                sorted(list(
                    set([e[-1].strip()
                         for e in list(g[1])
                         if e[-1].strip() and e[-1].strip() != "*"]))))
            # 使わない情報は取り除く。
            extra = re.sub(r"ドメイン:[^:\s]+", "", extra)
            extra = re.sub(r"[反同]義:[^\s]+", "", extra)
            extra = re.sub(r"カテゴリ:[^\s]+", "", extra)
            # 動詞の「付属情報候補」の情報については、重複エントリ
            # として複数のファイルにディスパッチ済み。ただこれだと
            # 後続の処理が区別出来ないので、区別出来るように書き換
            # えておく。
            if "付属動詞" in extra and hin != "動詞(付属動詞候補)":
                extra = re.sub(r"(付属動詞候補（[^（）]+）)", r"\1、の「自立」版", extra)
            yield entrycols + [extra]

    for hin, ents in sorted(all_entries.items()):
        ofn = posixpath.join(blddir, hinshi_mapping[hin] + ".csv")
        logging.info("dumping preprocessed output %r", ofn)
        cont = "\n".join([",".join(e) for e in _cleanup(hin, ents)]) + "\n"
        with io.open(ofn, "wb") as fo:
            fo.write(cont.encode("utf-8"))


# ==================================================================
# 第二段階関連のもろもろ
# ==================================================================
fieldnames_all = [
    "表層形",
    "左文脈ID", "右文脈ID",
    "コスト",
    "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3",
    "活用形1", "活用形2",
    "原形", "読み", "発音"
]
fieldnames_all_juman_1st = [
    # プリプロセス後の構造。元のとは違う。
    "代表表記",
    "原形",
    "表層形",
    "左文脈ID", "右文脈ID",
    "コスト",
    "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3",
    "読み", "出典等"
]
fieldnames_all_juman_2nd = [
    # プリプロセス後の構造。元のとは違う。
    "代表表記",
    "代表表記(読み)",
    "原形",
    "原形(読み)",
    "同一原形(読み)内順位",
    "表層形",
    "コスト",
    "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3",
    "活用形1", "活用形2", "-活用形2内部用-",
    "読み", "出典等"
]


# 第二段階と第四段階で使う。構造は最終形態のもの。
_omap = {
    ('形容詞',): (
        "Adj",
        ("表層形", "品詞細分類1", "活用形1", "活用形2", "読み",),
        ),
    ('連体詞',): (
        "Adnominal",
        ("表層形", "読み"),
        ),
    ('副詞',): (
        "Adverb",
        ("表層形", "品詞細分類1", "読み"),
        ),
    ('接続詞',): (
        "Conjunction",
        ("表層形", "品詞細分類1", "読み"),
        ),
    ('名詞', '一般',): (
        "Noun.general",
        ("表層形", "品詞細分類2", "読み"),
        ),
    ('名詞', '代名詞',): (
        "Noun.demonst",
        ("表層形", "品詞細分類2", "読み"),
        ),
    ('名詞', '副詞可能',): (
        "Noun.adverbal",
        ("表層形", "読み"),
        ),
    ('名詞', '形容動詞語幹',): (
        "Noun.adjv",
        ("表層形", "読み"),
        ),
    ('名詞', 'ナイ形容詞語幹',): (
        "Noun.nai",
        ("表層形", "読み"),
        ),
    ('名詞', '固有名詞', '人名', '名',): (
        "Noun.name.first",
        ("表層形", "品詞細分類3", "読み"),
        ),
    ('名詞', '固有名詞', '人名', '一般',): (
        "Noun.name.general",
        ("表層形", "品詞細分類3", "読み"),
        ),
    ('名詞', '固有名詞', '地域', '一般',): (
        "Noun.place.general",
        ("表層形", "品詞細分類3", "読み"),
        ),
    ('名詞', '固有名詞', '地域', '国',): (
        "Noun.place.country",
        ("表層形", "品詞細分類3", "読み"),
        ),
    ('名詞', '固有名詞', '組織',): (
        "Noun.org",
        ("表層形", "読み"),
        ),
    ('名詞', '固有名詞', '一般',): (
        "Noun.proper",
        ("表層形", "読み"),
        ),
    ('名詞', '数',): (
        'Noun.number',
        ("表層形", "読み"),
        ),
    ('名詞', '接尾',): (
        "Suffix",
        ("表層形", "品詞細分類2", "読み"),
        ),
    ('名詞', 'サ変接続',): (
        "Noun.verbal",
        ("表層形", "読み"),
        ),
    ('名詞',): (  # 非自立、特殊、動詞非自立的、etc.
        'Noun.others',
        ("表層形", "品詞細分類1", "品詞細分類2", "読み"),
        ),
    ('感動詞',): (
        "Interjection",
        ("表層形", "読み"),
        ),
    ('接頭詞',): (
        "Prefix",
        ("表層形", "品詞細分類1", "読み"),
        ),
    ('動詞',): (
        "Verb",
        ("表層形", "品詞細分類1", "活用形1", "活用形2", "原形", "読み"),
        ),
    ('記号',): (
        "Symbol",
        ("表層形", "品詞細分類1", "原形", "読み"),
        ),
    }

def _repr_entry(d, compact=True, exclude=[]):
    # デバッグ用。
    pkeys_full = [
        "代表表記",
        "原形",
        "表層形",
        "左文脈ID", "右文脈ID",
        "コスト",
        "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3",
        "活用形1", "活用形2",
        "読み", "出典等"
    ]
    exckifc = ("左文脈ID", "右文脈ID", "コスト")
    pkeys = []
    for k in [k for k in pkeys_full + list(d.keys()) if k not in exclude]:
        if k not in pkeys:
            if not compact or k not in exckifc:
                pkeys.append(k)
    return "-".join([d.get(k, " ") for k in pkeys
                     if not compact or d.get(k) != "*"])


def _repr_entries(ents, compact=True):
    # デバッグ用。
    return [_repr_entry(e, compact) for e in ents]


def _nop(s):
    return s


def _chop(s):
    return s[:-1]


class _sub(object):
    def __init__(self, rgx, repl):
        self._rgx = re.compile("({})".format("|".join([rgx, _h2k(rgx)])))
        self._repl = repl
    def __call__(self, s):
        return self._rgx.sub(self._repl, s)


class _returnitself(object):
    def __init__(self, s):
        self._s = s
    def __call__(self, yomi):
        return self._s

class __meireivar(object):
    def __call__(self, yomi):
        if yomi.endswith("ロ"):
            return "命令ｒｏ"
        elif yomi.endswith("ヨ"):
            return "命令ｙｏ"
        elif yomi.endswith("イ"):
            return "命令ｉ"
        return "命令ｅ"

_meireivar = __meireivar()

class _jdreader_filter(object):
    def __init__(self, jdent):
        jdent["読み"] = jdent["読み"]
        jdent.pop("左文脈ID")
        jdent.pop("右文脈ID")
        jdent["活用形1"] = "*"
        jdent["活用形2"] = "*"

        self._jdent = jdent

        self._sur = self._jdent["表層形"]
        self._origc = self._jdent["品詞"]
        self._origcd1 = self._jdent["品詞細分類1"]
        self._origcd2 = self._jdent["品詞細分類2"]
        self._origcd3 = self._jdent["品詞細分類3"]
        self._extra = self._jdent["出典等"]

    # 以下の動詞のに限らず活用語のマッピングは、正直言ってしまえば、
    # 語幹だけから自力生成の方が楽かもしれんのだが、まぁ一応「左から
    # 右へ」で頑張ってる。ゆえに「オカシナオカシナ移送」が多発して
    # 「ほんまにこれが楽なの?」的ではあるが、一応こっちのやり方にも
    # メリットはある。どういうことかというと、「活用形2」の主張だけ
    # 信じれば「特殊事情は jumandic の元エントリが頑張ってくれてる」
    # ということ。これでもまだわかりにくいと思うのでもう一言。
    # 「活用形1による特殊事情をあまり考えなくていい」てこと。
    # 「イ音便」「促音便」といった「活用形1のバリエーション」は
    # この移送方式では「ちょっとしか気にしなくていい」わけだ。
    # (これでも何を言ってるのかわかりにくければ、以下の _tmap の
    # キーと、利用している側の対応を見てほしい。「五段・カ行イ音便」
    # にマッピングする部分が _tmap の「五段」を使っている。自家発電
    # アプローチを採用する場合は、「五段・カ行イ音便」に相応しい
    # 活用を個々に書く必要があるところである。)
    def _kmapping(self, hd1, kk1):
        _tmap = {  # 一番多いものがベース。
            "一段": {
                "未然形": (_returnitself("未然形"), _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),  # 代用(目的外使用)
                "基本連用形": (_returnitself("連用形"), _nop),
                "文語命令形": (_meireivar, _nop),
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                "タ接連用形": (
                    _returnitself("体言接続特殊"),
                    _sub(r"$", "ん")),  # 代用(目的外使用)
                },
            "五段": {
                "未然形": (_returnitself("未然形"), _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),  # 代用(目的外使用)
                "基本連用形": (_returnitself("連用形"), _nop),
                "タ接連用形": (_returnitself("連用タ接続"), _nop),
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                },
            "五段・サ行": {
                "未然形": (_returnitself("未然形"), _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),  # 代用(目的外使用)
                "基本連用形": (_returnitself("連用形"), _nop),
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                },
            "五段・ワ行促音便": {
                "未然形": (_returnitself("未然形"), _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),  # 代用(目的外使用)
                "基本連用形": (_returnitself("連用形"), _nop),
                "タ接連用形": (_returnitself("連用タ接続"), _nop),
                "命令形": (_meireivar, _nop),
                },
            "五段・ラ行": {
                "未然形": (_returnitself("未然形"), _nop),
                "基本連用形": (_returnitself("連用形"), _nop),
                "タ接連用形": (_returnitself("連用タ接続"), _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                "命令形": (_meireivar, _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),
                "タ系省略推量形": (
                    _returnitself("未然特殊"),
                    _sub(r"っ?たろ$", "ん")),  # 代用(目的外使用)
                "タ形": (
                    _returnitself("体言接続特殊"),
                    _sub(r"っ?た$", "ん")),  # 代用(目的外使用)
                "タ系推量形": (
                    _returnitself("体言接続特殊２"),
                    _sub(r"っ?たろう$", "")),  # 代用(目的外使用)
                },
            "五段・ラ行特殊": {
                "未然形": (_returnitself("未然形"), _nop),
                "基本連用形": (_returnitself("連用形"), _nop),
                "タ接連用形": (_returnitself("連用タ接続"), _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                "命令形": (_meireivar, _nop),
                "文語命令形": (_meireivar, _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),
                "タ形": (
                    _returnitself("未然特殊"),
                    _sub(r"っ?た$", "ん")),  # 代用(目的外使用)
                },
            "五段・ワ行": {  # 促音便とウ音便だが、処理上は同じでいい。
                "未然形": (_returnitself("未然形"), _nop),
                "基本連用形": (_returnitself("連用形"), _nop),
                "タ接連用形": (_returnitself("連用タ接続"), _nop),
                "命令形": (_meireivar, _nop),
                "意志形": (_returnitself("未然ウ接続"), _chop),
                },
            #"カ変・クル": {
            #     "未然形": (_returnitself("未然形"), _nop),
            #     "基本連用形": (_returnitself("連用形"), _nop),
            #     "音便条件形": (_returnitself("仮定縮約１"), _nop),
            #     "命令形": (_meireivar, _nop),
            #     "文語命令形": (_meireivar, _nop),
            #     "意志形": (_returnitself("未然ウ接続"), _chop),
            #     "タ形": (
            #         _returnitself("体言接続特殊"),
            #         _sub(r"っ?た$", "ん")),  # 代用(目的外使用)
            #     "タ系推量形": (
            #         _returnitself("体言接続特殊２"),
            #         _sub(r"っ?たろう$", "")),  # 代用(目的外使用)
            #     },
            "サ変・\uff0dスル": {
                "文語基本形": (_returnitself("文語基本形"), _nop),
                "タ接連用形": (_returnitself("未然形"), _nop),  # 代用(目的外使用)
                "基本連用形": (
                    _returnitself("未然ウ接続"),
                    _sub(r"$", "よ")),  # 代用(目的外使用)
                "意志形": (
                    _returnitself("未然ウ接続(ょ)"),
                    _sub(r"よう$", "ょ")),  # 代用(目的外使用)
                "文語未然形": (_returnitself("未然レル接続"), _nop),  # 代用(目的外使用)
                "文語命令形": (_meireivar, _nop),
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                },
            "サ変・\uff0dズル": {
                "文語基本形": (_returnitself("文語基本形"), _nop),
                "文語未然形": (_returnitself("未然形"), _nop),
                "文語命令形": (
                    _returnitself("未然ウ接続"),
                    _sub(r"じろ$", "ぜよ")),  # 代用(目的外使用)
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                },
            }
        for k in _tmap.keys():
            _tmap[k]["基本形"] = (_returnitself("基本形"), _nop)
            _tmap[k]["基本条件形"] = (_returnitself("仮定形"), _chop)
        return _tmap[kk1]

    def _verb(self):
        # 「接尾/非自立のみ」とする、「接尾/非自立も」とする、の判断
        # 基準はかなり曖昧で、ipadic でも悩みつつ決めてるんじゃないか
        # と思う。jumandic が迷い、悩んでいるのは補足列ではっきり表明
        # されている。そして私にとってもこれは同じ。たとえば「いらっ
        # しゃる」は、ipadic では自立版と非自立版の両方が定義されてい
        # る。対して「らっしゃる」は非自立版のみが定義されている。一見
        # するとこれはあたかも自明のように感じられるが、それは相手が
        # 「(い)らっしゃる」だからである。ものによってはどちらともいえ
        # ないものが多い。「らっしゃる」に関しては jumandic は自立動詞
        # としても考えている。んなわけあるかいな、てほどではない、これ
        # は。八百屋のおやっさんが「らっしゃい」言うのは違うんかい、
        # てハナシ。逆もあって、「自立でしかあるまいよ」と思っても、
        # たとえば「おっしゃる」は普通は自立と解釈するであろうが、
        # 「うそ-おっしゃい」のようなケースで、非自立が接続していると
        # 考えることにも一理ある。一般的に言って「自立または非自立一方
        # としてしか存在し得ない動詞」てのは、対象がカジュアルになれば
        # なるほど成立しなくなってくる。「非自立/接尾」は入れておくと
        # 便利なことが結構多いと思うわけだが、その際に「非自立しかいな
        # いから困る」よりは、両方いればいいのかな、と思うことにした。
        # ゆえに ipadic が「これは非自立でしかありえんだろう」と考えて
        # いたとしても、jumandic が自立としても考えているならそっちに
        # 従おうと。(幸い「非自立,五段・ラ行特殊」は制約がないのでこう
        # いうことが出来る。)処理上は、プリプロセスで「自立用/非自立用」
        # に重複ディスパッチしているので、その際に埋めた補足情報に基く。
        if "付属動詞" in self._extra and "「自立」版" not in self._extra:
            return self._verbsuf()

        self._jdent["品詞"] = "動詞"
        self._jdent["品詞細分類1"] = "自立"
        self._jdent["品詞細分類2"] = "*"
        self._jdent["品詞細分類3"] = "*"
        targm = {}
        if self._origcd2 == "母音動詞":
            # プリプロセスでの「一段・クレル」「一段・得る」についての
            # コメント参照。ここに至っては、この「母音動詞なるものの
            # 実力とやら」は全部「一段」でいい、と思う。にしても
            # この「母音動詞」って、何が、どこが母音なのだろうか?
            self._jdent["活用形1"] = "一段"
            targm = self._kmapping("自立", self._jdent["活用形1"])
        elif self._origcd2 in ("サ変動詞", "ザ変動詞"):
            # 「サ変・スル」相当はプリプロセスで除外済み。
            if self._origcd2 == "ザ変動詞":
                self._jdent["活用形1"] = "サ変・\uff0dズル"
            else:
                self._jdent["活用形1"] = "サ変・\uff0dスル"
            targm = self._kmapping("自立", self._jdent["活用形1"])
        elif self._origcd2 == "子音動詞カ行":
            # カ行はイ音便と促音便の二つである。
            # 後者はプリプロセスで除外済み(そちらのコメント
            # 参照)。
            self._jdent["活用形1"] = "五段・カ行イ音便"
            targm = self._kmapping("自立", "五段")
        elif self._origcd2 == "子音動詞ガ行":
            self._jdent["活用形1"] = "五段・ガ行"
            targm = self._kmapping("自立", "五段")
        elif self._origcd2 == "子音動詞サ行":
            self._jdent["活用形1"] = "五段・サ行"
            targm = self._kmapping("自立", "五段・サ行")
        elif self._origcd2 == "子音動詞タ行":
            self._jdent["活用形1"] = "五段・タ行"
            targm = self._kmapping("自立", "五段")
        elif self._origcd2 == "子音動詞ナ行":
            self._jdent["活用形1"] = "五段・ナ行"
            targm = self._kmapping("自立", "五段")
        elif self._origcd2 == "子音動詞バ行":
            self._jdent["活用形1"] = "五段・バ行"
            targm = self._kmapping("自立", "五段")
        elif self._origcd2 == "子音動詞マ行":
            self._jdent["活用形1"] = "五段・マ行"
            targm = self._kmapping("自立", "五段")

        #
        # 五段・ラ行と五段・ラ行特殊の関係に関しては、
        # ipadic の考え方が正しいとは思わない。これは
        # 例えば「下さる」で連用形が「下さり-」になるか
        # 「下さい-」になるかである。一応体言接続特殊、
        # 体言接続特殊２の有無の差もあるが、本質的には
        # この連用形の「い」になるか「り」になるかの
        # 区別である。jumandic でも「ラ行イ形」として
        # 区別しているので、同罪ではある。これは定義の
        # 設計ミスか、あるいは「両方入れときゃいいのに」
        # のどちらかである。「設計ミス」と言っているのは、
        # これは「活用形1」で区分するものではなくて、
        # 「活用形2」で区別すべきものではないのかという
        # こと。実際のところは、シード辞書・コーパスから
        # 始めてトレーニングまで込みで設計なので、活用形1
        # で区別する意義があるのかもしれないけれど、
        # エンドユーザ目線ではこれはどちらにしてもかなり
        # 迷惑な話。
        # ひとまずこの読み替えフェーズではそのまま jumandic
        # の主張を信じておいて、後続の _accept でこれを
        # 同じものとみなしてハジくことにする。
        #
        elif self._origcd2 == "子音動詞ラ行":
            self._jdent["活用形1"] = "五段・ラ行"
            targm = self._kmapping("自立", "五段・ラ行")
        elif self._origcd2 == "子音動詞ラ行イ形":
            self._jdent["活用形1"] = "五段・ラ行特殊"
            targm = self._kmapping("自立", "五段・ラ行特殊")

        elif self._origcd2 == "子音動詞ワ行":
            self._jdent["活用形1"] = "五段・ワ行促音便"
            targm = self._kmapping("自立", "五段・ワ行")
        elif self._origcd2 == "子音動詞ワ行文語音便形":
            self._jdent["活用形1"] = "五段・ワ行ウ音便"
            targm = self._kmapping("自立", "五段・ワ行")
        else:
            return False
        if self._origcd3 not in targm:
            return False
        self._jdent["活用形2"] = targm[self._origcd3][0](self._jdent["読み"])
        self._jdent["表層形"] = targm[self._origcd3][1](self._jdent["表層形"])
        self._jdent["読み"] = _h2k(targm[self._origcd3][1](self._jdent["読み"]))
        return True

    def _verbsuf(self):
        self._jdent["品詞"] = "動詞"
        self._jdent["品詞細分類2"] = "*"
        self._jdent["品詞細分類3"] = "*"
        self._jdent["活用形2"] = self._origcd3
        # ------
        _tmap = {
            "母音動詞": (
                "接尾", "一段", self._kmapping("接尾", "一段"),
                ),
            #"子音動詞サ行": (
            #    "接尾", "五段・サ行", self._kmapping("接尾", "五段・サ行"),
            #    ),
            "子音動詞ラ行": (
                "接尾", "五段・ラ行", self._kmapping("接尾", "五段・ラ行"),
                ),
            #"カ変動詞": (
            #    "非自立", "カ変・クル", self._kmapping("非自立", "カ変・クル"),
            #    ),

            "子音動詞カ行": (
                # 「おく」の対応より。
                "非自立", "五段・カ行イ音便", self._kmapping("非自立", "五段"),
                ),
            "子音動詞カ行促音便形": (
                # "ユク"があれば「五段・カ行促音便ユク」だったとこころ。
                # たまたま jumandic では入れてない。
                "非自立", "五段・カ行促音便", self._kmapping("非自立", "五段"),
                ),
            "子音動詞ラ行イ形": (
                "非自立", "五段・ラ行特殊", self._kmapping("非自立", "五段・ラ行特殊"),
                ),
            "子音動詞ワ行": (
                # 「しまう」の対応より。
                "非自立", "五段・ワ行促音便", self._kmapping("非自立", "五段・ワ行促音便"),
                ),
            # 子音動詞マ行:
            # 「あぐむ」のみなのだが、非自立の五段・マ行は「込む」
            # のためだけのものなので、これには使えない。(動詞,接尾
            # にはマ行がないのでますます使えない。)これは自立動詞
            # として ipadic に既存なので、捨ててしまうことにする。
            #
            }
        if self._origcd2 not in _tmap:
            #logging.info("%r, %r, %r, %r", self._sur, self._origcd1, self._origcd2, self._origcd3) #, self._extra)
            return False
        targm = _tmap[self._origcd2]
        self._jdent["品詞細分類1"] = targm[0]
        self._jdent["活用形1"] = targm[1]
        if self._origcd3 not in targm[2]:
            #if "タ接" not in self._origcd3:
            #logging.info("%r, %r, %r, %r", self._sur, self._origcd1, self._origcd2, self._origcd3) #, self._extra)
            return False
        self._jdent["活用形2"] = targm[2][self._origcd3][0](self._jdent["読み"])
        self._jdent["表層形"] = targm[2][self._origcd3][1](self._jdent["表層形"])
        self._jdent["読み"] = _h2k(targm[2][self._origcd3][1](self._jdent["読み"]))
        return True

    def _adj_var(self):
        self._jdent["品詞"] = "形容詞"
        self._jdent["品詞細分類2"] = "*"
        self._jdent["品詞細分類3"] = "*"
        self._jdent["活用形2"] = self._origcd3
        # ------
        _tmap = {
            "自立": {
                "基本形": (_returnitself("基本形"), _nop),
                "文語基本形": (_returnitself("文語基本形"), _nop),
                "文語未然形": (_returnitself("未然ヌ接続"), _nop),
                "基本省略推量形": (_returnitself("未然ウ接続"), _nop),  # 代用(目的外使用)
                "基本推量形": (
                    _returnitself("連用タ接続"),
                    _sub(r"ろう$", "っ")),  # 代用(目的外使用)
                "基本連用形": (_returnitself("連用テ接続"), _nop),
                "文語命令形": (
                    _returnitself("連用テ接続(っ)"),
                    _sub(r"かれ$", "くっ")),  # 代用(目的外使用)

                # ipadicの「ゴザイ接続」ってちょっと腑に落ちなくてな。
                # 「いとしぅございます」はいい。けれど「危なげなう
                # ございます」は「危なげのう」ではないのか。「なう」
                # の「のう」化、「たう」の「とう」化はやってないんだわ。
                # なんでだろ。純粋に古文相手の場合はいいんだけれど、
                # 「現代に再現されている時代劇調のドラマ」で頻出
                # するわけよね、これら。現代人は「危なげのう」と
                # 書くだろう。せっかくこういうのを定義したんだったら
                # そこまでやれば良かったのに。私がここで独自にそう
                # してもいいんだけれど、「ゴーっにいっては業に従う」
                # のがこの「スクリプトを主張するうんちゃらかんちゃら」
                # のコンセプトなので…、まぁ ipadic と同じにしとくわ。
                "文語連用形": (_returnitself("連用ゴザイ接続"), _nop),
                "タ形": (
                    _returnitself("連用ゴザイ接続(ぅ)"),
                    _sub(r"(かっ)?た$", "ぅ")),  # 代用(目的外使用)

                "文語連体形": (_returnitself("体言接続"), _nop),
                "基本条件形": (_returnitself("仮定形"), _chop),
                "命令形": (_meireivar, _nop),
                "音便条件形": (_returnitself("仮定縮約１"), _nop),
                "音便条件形２": (_returnitself("仮定縮約２"), _nop),
                "語幹": (_returnitself("ガル接続"), _nop),  # 代用(目的外使用)
                },
        }
        if self._origcd2 not in ("イ形容詞アウオ段", "イ形容詞イ段"):
            return False
        targm = _tmap["自立"]
        self._jdent["品詞細分類1"] = "自立"
        self._jdent["活用形1"] = self._origcd2.replace("イ形容詞", "形容詞・")
        if self._origcd3 not in targm:
            return False
        self._jdent["活用形2"] = targm[self._origcd3][0](self._jdent["読み"])
        self._jdent["表層形"] = targm[self._origcd3][1](self._jdent["表層形"])
        self._jdent["読み"] = _h2k(targm[self._origcd3][1](self._jdent["読み"]))
        return True

    def _adj(self):
        if self._origcd2 in ("ナ形容詞", "ナノ形容詞", "ナ形容詞特殊"):
            # これは形容詞ではない、 ipadic 流儀なら。
            # ipadic にあるものでは「きらびやか」「プラトニック」
            # など。jumandic で「ナ形容詞」「ナノ形容詞」と言ってる
            # もの全てがこれ。なお、プリプロセスで「語幹」だけに
            # してるので、活用形の方は気にしなくていい、ここでは。
            # (「きらびやかだったろう」とか、なんだよそれ、と。)
            self._jdent["品詞"] = "名詞"
            self._jdent["品詞細分類1"] = "形容動詞語幹"
            self._jdent["品詞細分類2"] = "*"
            self._jdent["品詞細分類3"] = "*"
            self._jdent["活用形1"] = "*"
            self._jdent["活用形2"] = "*"
            self._jdent["原形"] = self._jdent["表層形"]
            self._jdent["代表表記"] = re.sub(r"[なだ]$", "", self._jdent["代表表記"])

        elif self._origcd2 in ("タル形容詞",):
            # Wikipedia「形容動詞」説明中の「タルト型活用」と、その
            # 「学校文法の口語文法では、連用形は副詞、連体形は連体詞
            # として扱われる。」を ipadic は忠実にやろうとしている
            # 片鱗はある。必ずしも一貫性はないものの、「雑然」「殺伐」
            # の定義を観察するとわかる。
            # jumandic の「タル形容詞」はこの「タルト型」とは若干違う
            # ものの、違うのは考え方だけで中身は同じなので、この
            # マッピング目的に使える。(jumandic 「タル形容詞」が「基
            # 本形」と称しているものが、タルト活用の連体形である。)
            self._jdent["原形"] = self._jdent["表層形"]
            if self._origcd3 == "語幹":  # 溌剌
                self._jdent["品詞"] = "名詞"
                self._jdent["品詞細分類1"] = "形容動詞語幹"
                self._jdent["代表表記"] = re.sub(
                    r"たる$", "", self._jdent["代表表記"])
            elif self._origcd3 == "基本連用形":  # 溌剌と
                self._jdent["品詞"] = "副詞"
                self._jdent["品詞細分類1"] = "一般"
                self._jdent["代表表記"] = re.sub(
                    r"たる$", "と", self._jdent["代表表記"])
            elif self._origcd3 == "基本形":  # 溌剌たる
                self._jdent["品詞"] = "連体詞"
                self._jdent["品詞細分類1"] = "*"
            self._jdent["品詞細分類2"] = "*"
            self._jdent["品詞細分類3"] = "*"
            self._jdent["活用形1"] = "*"
            self._jdent["活用形2"] = "*"

        elif "(:→名詞/ナイ形容詞語幹:)" in self._extra:
            # _rgx_adj_nai のコメント参照。
            # というか、なので「どこが名詞,ナイ形容詞語幹やねん」と思うなら
            # それは jumandic 作者の責任ではなくワタシのものだぴょーん、許して
            # ちょんまげ。
            self._jdent["品詞"] = "名詞"
            self._jdent["品詞細分類1"] = "ナイ形容詞語幹"
            self._jdent["品詞細分類2"] = "*"
            self._jdent["品詞細分類3"] = "*"
            self._jdent["活用形1"] = "*"
            self._jdent["活用形2"] = "*"

            # こいつに関しては、表層形なども全部変える必要がある。
            # 少しだけめんどい。変更の対象は:
            # ・代表表記
            # ・表層形
            # ・原形
            # ・読み
            # の4つ。幸いこいつに関しては「因がおう報」パターンは邪魔にならない。
            of = self._jdent["代表表記"]
            base, nai = _rgx_adj_nai.match(of).group(1, 2)
            self._jdent["代表表記"] = base
            self._sur = self._jdent["表層形"] = self._jdent["表層形"][:-len(nai)]
            self._jdent["原形"] = self._jdent["原形"][:-len(nai)]
            self._jdent["読み"] = self._jdent["読み"][:-len(nai)]

        elif not self._adj_var():
            return False
        return True

    def _special(self):
        self._jdent["品詞"] = "記号"
        if self._origcd1 in ("句点", "読点", "空白"):
            self._jdent["品詞細分類1"] = self._origcd1
        elif self._origcd1 == "括弧始":
            self._jdent["品詞細分類1"] = "括弧開"
        elif self._origcd1 == "括弧終":
            self._jdent["品詞細分類1"] = "括弧閉"
        elif self._origcd1 == "記号":
            # プリプロセスで捨てたもの以外は「記号,一般」でいい。
            # ただ、「顔文字」は区別ないと時として迷惑なので、
            # コメントにその旨追加する。(「）：」が顔文字に
            # 入っている。こういう、いくつかフォーマルな文書で
            # 顔文字以外で使うものが入っている。
            self._jdent["品詞細分類1"] = "一般"

    def _demonstrative(self):
        # ipadic の言うところの demonstrative とは少々違う。
        # ipadic ではこれは「代名詞」だが、jumandic でのこれは、
        # ipadic での代名詞と連体詞と副詞に相当するものがいっしょ
        # になってる。
        if self._origcd1 == "名詞形態指示詞":
            self._jdent["品詞"] = "名詞"
            self._jdent["品詞細分類1"] = "代名詞"
            # jumandic では「一般」と「縮約」の区別はしてないし、
            # 「縮約」相当のエントリはない。
            self._jdent["品詞細分類2"] = "一般"
        elif self._origcd1 == "連体詞形態指示詞":
            self._jdent["品詞"] = "連体詞"
            self._jdent["品詞細分類1"] = "*"
        elif self._origcd1 == "副詞形態指示詞":
            self._jdent["品詞"] = "副詞"
            # ipadic での「助詞類接続」は「淡々」「ニヤリ」など
            # 「～とする/なる」のようになるやつらで、 jumandic
            # ではこれを区別してない。(「副詞」に入るものたち
            # 一般の話は「副詞」の話のなかで。)
            self._jdent["品詞細分類1"] = "一般"

    def _noun_suusi(self):
        # Noun.suusi のものは、 ipadic で「名詞,数」のものと
        # 「名詞,接尾,助数詞」「接頭詞,数接続」のものが区別されずに
        # 混在している。これをシンプルなルールで振り分けることは
        # 出来ないので、基本的に「おばかな」方法で分類する。
        # (なお、ipadic の Noun.number でも「名詞,接尾,助数詞」
        # 「接頭詞,数接続」相当のもの(「幾」や「何」)が含まれるの
        # だが、さすがに単位に相当するもの(「キロ」など)は区別
        # され、Suffix にいる。)

        # 「ミリ」「ミリオン」などと「分の」。
        if re.match(r"^[キギテミメ]", self._sur) or self._sur == "分の":
            self._jdent["品詞"] = "名詞"
            self._jdent["品詞細分類1"] = "接尾"
            self._jdent["品詞細分類2"] = "助数詞"
        else:
            # 「接頭詞,数接続」に関しては ipadic でもちゃんと
            # してないので、以外は「名詞,数」にする。
            self._jdent["品詞"] = "名詞"
            self._jdent["品詞細分類1"] = "数"
            self._jdent["品詞細分類2"] = "*"

    def _adverb(self):
        # ipadic には「副詞,一般」と「副詞,助詞類接続」の区別があるが、
        # jumandic の方にはない。かろうじてレコードの最後の列(self._extra)
        # に「数量修飾」などの区分がところどころ入っているが
        # 「助詞類接続」かどうかに使えるものではなさそう。なのでもう
        # 一括で「副詞,一般」とする。
        # このように、ここでの読み替えは至ってシンプルで、何も問題なさそうに
        # 思えるのだけれど、よくよく観察すると結構おかしくって、「あるいは」
        # だの「かくして」だのが副詞になってしまっている。一般にはこれは
        # 接続詞であるし、ipadic でもそうしている。
        # jumandic 固有追加のものについてはどうしようもないけれど、ipadic
        # に既存のものに関しては、後続の合致判定で接続詞と同一視することで
        # ハジく。
        self._jdent["品詞細分類1"] = "一般"

    def _noun_keishiki(self):
        # Noun.keishiki のものは ipadic の「Noun.others」相当のものだが、
        # ipadicでの以下:
        #    名詞,動詞非自立的
        #    名詞,特殊,助動詞語幹
        #    名詞,非自立,一般
        #    名詞,非自立,形容動詞語幹
        #    名詞,非自立,助動詞語幹
        #    名詞,非自立,副詞可能
        # に該当する区分はなく全て「形式名詞」。
        # 全部「名詞,非自立,一般」にしとく。
        self._jdent["品詞"] = "名詞"
        self._jdent["品詞細分類1"] = "非自立"
        self._jdent["品詞細分類2"] = "一般"

    def _suffix(self):
        # jumandic ではなんだか「接尾」にやたら力を入れている。
        # 気持ちはわからないでもない。カジュアルな文章ほど
        # 妙な語尾になりがちだから。ただ「綺麗な設計」とは
        # 思わない。少なくとも見覚えのない「理論」で、いわゆる
        # 「がっこでならった」のとは随分違って、何か一般教養
        # レベル以上の学説があるんだろうか。独自の考え方を
        # 使っている。けど、正直いって ipadic の「がっこで
        # ならったふつーのやつ」の方が柔軟で綺麗だと思う。
        # で、 jumandic ソースでの「Suffix」に収められている
        # 接尾は、ここでの読み替えによって色んなものに変える
        # ので注意。(ipadic の「接尾」と「非自立」は使いにくい
        # という実現可能性上の理由もあるが、シンプルに「接尾
        # でなくてもええやん」というのがかなりあるので。)

        self._jdent["品詞"] = "名詞"
        self._jdent["品詞細分類1"] = "接尾"
        if self._origcd1 == "名詞性名詞助数辞":
            self._jdent["品詞細分類2"] = "助数詞"
        elif self._origcd1 == "名詞性特殊接尾辞":
            # 「以降」とか「末」とか。ただ、ipadic で「地域」扱いのものまで
            # なぜかここにいるので、それらだけ別個に「地域」に。(「省」だけ
            # は微妙だ。これは「環境省」みたいなのもあるわけで。)
            if self._sur in ("都", "府", "県", "市", "村", "州", "省",):
                self._jdent["品詞細分類2"] = "地域"
            else:
                self._jdent["品詞細分類2"] = "副詞可能"
        elif self._origcd1 == "名詞性述語接尾辞":
            # 「っぷり」とか「放題」
            self._jdent["品詞細分類2"] = "形容動詞語幹"

        elif self._origcd1 == "名詞性名詞接尾辞":
            self._jdent["品詞細分類2"] = "一般"
            # この時点では区別しない。後続の処理で、ipadic との合致をみて
            # 調整される。

        elif re.match("形容詞性.*接尾辞", self._origcd1):
            if self._origcd2 in ("ナノ形容詞", "ナ形容詞"):
                self._jdent["品詞細分類2"] = "一般" if self._origcd1 == "形容詞性名詞接尾辞" else "形容動詞語幹"
                self._jdent["品詞細分類3"] = "*"
                self._jdent["原形"] = self._jdent["表層形"]
                self._jdent["代表表記"] = re.sub(r"[なだ]$", "", self._jdent["代表表記"])
            else:
                # 「形容詞性述語接尾辞,イ形容詞アウオ段」のみ。
                # 「形容詞性名詞接尾辞,イ形容詞アウオ段」はプリプロセスで
                # 除外してある。jumandic はなんでも形容詞にしたがる傾向が
                # あるけれど、こいつに関しては逆で、ipadic だと「形容詞,
                # 接尾」「形容詞,非自立」のパターン。
                if not self._adj_var():
                    return False
                # ここで「接尾」とするんだけれど、これは無論 jumandic が
                # この区別を持ってないから仕方なく、てこと。本来は ipadic
                # は非自立と接尾を区別してる。(後続の合致判定の部分で同一
                # 視する。)
                self._jdent["品詞細分類1"] = "接尾"
                #logging.info(_repr_entry(self._jdent))
        elif self._origcd1 == "動詞性接尾辞":
            if not self._verbsuf():
                return False
        return True

    def _prefix(self):
        self._jdent["品詞"] = "接頭詞"
        self._jdent["品詞細分類2"] = "*"
        # 「動詞接頭辞」→「動詞接続」、「名詞接頭辞」→「名詞接続」、
        # と、ここまではいいのだが、ipadic の残りは「形容詞接続」
        # と「数接続」、jumandic の残りは「イ形容詞接頭辞」と
        # 「ナ形容詞接頭辞」。どうせいつものごとく「ナ形容詞が悪
        # なのであろう」かというと、ここではそうじゃない。「ナ形容詞」
        # は要するに形容動詞語幹名詞なので、素直に名詞接続にマッピング
        # 出来る。けれども、「数接続」は機械判定では決定できない。
        # ただこれに関しては、ipadic でも「ｐＨ」みたいに数専用でない
        # 限りは名詞接続と重複定義しているので、ここでは気にしない
        # ことにする。
        if self._origcd1 == "動詞接頭辞":
            self._jdent["品詞細分類1"] = "動詞接続"
        elif self._origcd1 in ("名詞接頭辞", "ナ形容詞接頭辞"):
            self._jdent["品詞細分類1"] = "名詞接続"
        else:
            self._jdent["品詞細分類1"] = "形容詞接続"
        return True

    def _noun_name(self):
        self._jdent["品詞細分類1"] = "固有名詞"
        self._jdent["品詞細分類2"] = "人名"
        if "人名:外国" in self._extra:
            # 外国人名はファースト・ラストの扱いが厄介なので、たとえ
            # それがファーストネームであろうがラストネームであろうが
            # 「一般」としてしまう。(著名人以外は必ずしも自明ではない。)
            self._jdent["品詞細分類3"] = "一般"
        else:
            # プリプロセスで捨ててるので残りは「名」である。
            self._jdent["品詞細分類3"] = "名"

    def _noun_place(self):
        # 「固有名詞,地名」はマジメに考え出すとそこそこややこしい、
        # のだが、これは「地名」という日本語のせい。ipadic で
        # Noun.place に収められているその英語で考えれば、もう少し
        # 意味が広がる(つまり「「場所」にまつわる固有名詞」)。なので、
        # 「環状八号線」のようなものを Noun.place と考える手はある。
        # けれど ipadic はそうせずにこれを一般の固有名詞としている。
        # jumandic はこの広い方の考えらしい。どちらの考えが正しい
        # とかはないけれど、狭い意味での「地名」だけでも膨大に
        # なりうるので、個人的には「環状八号線」は「固有名詞,一般」、
        # 「東名」は「固有名詞,地域」もしくは「名詞,一般」かなぁと
        # 思う。
        self._jdent["品詞細分類1"] = "固有名詞"
        self._jdent["品詞細分類2"] = "地域"
        self._jdent["品詞細分類3"] = "一般"
        if self._extra.endswith(":施設"):
            # 「place」として扱われた上での補足として「施設」といっている
            # だけなのだが、なんというかチグハグ。「ホワイトハウス」は
            # 施設だろうが地域ではなかろう。「チェルノブイリ」は「原発」
            # まで付かないなら施設ではなかろう。「東名」「関越」「名神」
            # は確かに地域だが、これもチェルノブイリと同じく。数も
            # 少ないので、オレ流に全部分類し直す。(結果 ipadic に既存
            # のエントリと同じになるものもあろう。)
            if self._sur in ("東名", "関越", "名神"):
                # 「施設」とみなすのは異論あるけれど、「固有名詞,地域」
                # なのはいいと思う。ゆえにこれはそのまま。
                pass
            elif self._sur in ("チェルノブイリ", "グリニッジ",):
                # 「施設」とみなすのは異論ある。ともに普通に地名。
                # 結果としてはそのまま。
                pass
            elif self._sur in ("バッキンガム", "クレムリン", "ホワイトハウス",):
                # この3つは共通して「王室や政府などの象徴」であり、
                # そういう理由でわざわざ入れたんだろう。ただこれら
                # 3つは個々に個々、かなぁと思う。
                if self._sur == "バッキンガム":
                    # バッキンガム宮殿の存在から「地名,施設」扱いなのだろう
                    # が、「バッキンガム公の宮殿」が始まりの宮殿なのだから、
                    # 「バッキンガム」だけを「place」扱いにするのは違うと思う。
                    # うーん、人名? かなぁ…。
                    self._jdent["品詞細分類2"] = "人名"
                    self._jdent["品詞細分類3"] = "一般"
                elif self._sur == "クレムリン":
                    # 旧ロシア帝国の宮殿にしてロシア政府の代名詞、
                    # ということであって、色んな使われ方をするので、
                    # 普通の一般の固有名詞でいいんじゃないかなと
                    # 思う。(ちなみに世界遺産登録されてるので、
                    # 確かに地名と考える考え方もめちゃくちゃ一般的。)
                    self._jdent["品詞細分類2"] = "一般"
                    self._jdent["品詞細分類3"] = "*"
                elif self._sur == "ホワイトハウス":
                    # ipadic 流儀の場合は一般固有名詞で、私もそれで
                    # いいと思う。
                    self._jdent["品詞細分類2"] = "一般"
                    self._jdent["品詞細分類3"] = "*"
            elif self._sur in ("マチュピチュ", "天安門",):
                # ともに、「固有名詞であり地名でもある」パターンかと。
                # ひとまずは「一般固有名詞」にさえいればいいかな?
                self._jdent["品詞細分類2"] = "一般"
                self._jdent["品詞細分類3"] = "*"
        elif self._extra.endswith("地名:国") or self._extra.endswith("地名:国:旧称"):
            self._jdent["品詞細分類3"] = "国"

    def _noun_except_proper(self):
        if self._origcd1 in ("普通名詞", "サ変名詞"):
            # シンプルに読み替えるだけでいい、これは。
            self._jdent["品詞細分類1"] = "一般" if self._origcd1 == "普通名詞" else "サ変接続"

        elif self._origcd1 == "時相名詞":
            # うーん。「別れぎわ」とか。副詞可能。
            self._jdent["品詞細分類1"] = "副詞可能"

        elif self._origcd1 == "副詞的名詞":
            # Noun.hukusi のものはストレートにマッピングできる。
            self._jdent["品詞細分類1"] = "副詞可能"

        elif self._origcd1 == "形式名詞":
            self._noun_keishiki()

    def __call__(self):
        self._jdent["-活用形2内部用-"] = ""  # 処理の都合上必要。

        if self._origc == "動詞":
            if not self._verb():
                return

        if self._origc == "特殊":
            self._special()

        elif self._origc == "副詞":
            self._adverb()

        elif self._origc == "指示詞":
            self._demonstrative()

        ###############################################
        # 「名詞」のうち、特殊、というか「あんまし頭使わずに
        # マッピング出来るものたち。
        elif self._origcd1 == "人名":
            self._noun_name()

        elif self._origcd1 == "地名":
            self._noun_place()

        elif self._origcd1 == "組織名":
            #logging.info("%r, %r", self._jdent, self._extra)
            self._jdent["品詞細分類1"] = "固有名詞"
            self._jdent["品詞細分類2"] = "組織"

        elif self._origcd1 == "固有名詞":
            #logging.info("%r, %r", self._jdent, self._extra)
            self._jdent["品詞細分類2"] = "一般"

        ###############################################
        # 「名詞」のうち、繊細な議論を要するものたち。
        elif "名詞" in self._origcd1 and "接" not in self._origc:
            self._noun_except_proper()
        ###############################################

        elif self._origcd1 == "数詞":
            self._noun_suusi()

        elif self._origc == "接尾辞":
            if not self._suffix():
                return

        elif self._origc == "接頭辞":
            if not self._prefix():
                return

        elif self._origc == "形容詞":
            if not self._adj():
                return

        elif self._origcd2 in (
            "ナ形容詞", "ナノ形容詞"):
            # 助動詞に「ナ形容詞」「ナノ形容詞」がいるが、冒頭で
            # 宣言した通り、助動詞は扱わないのでもーまんたい。
            return

        if "(" in self._jdent["活用形2"]:
            self._jdent["-活用形2内部用-"] = self._jdent["活用形2"]
            self._jdent["活用形2"] = re.sub(r"\(.*\)", "", self._jdent["活用形2"])

        yield self._jdent


def _secondstep(args):
    #
    # 何度も同じようなことばかりして非効率だ、と言わないでくれ。
    # 品詞等々もろもろ丸っきり別物に変えうる関係で、ipadic 流儀への
    # (仮)読み替え後に、再度のまとめ直しがどうしても必要なのである。
    #
    # ここでは第三段階以降がグルーピングを行なうのに相応しい形にソート
    # するのが主眼だが、これと同時に、「代表表記と原形の読み」と、
    # 「同一原形(読み)内順位」を追加する。
    #
    # 前者と後者は内容的にも処理的にも一体だが、目的は少し違う。
    #
    # 前者は、jumandic のまとめ方と私のやつのコンセプトがちょっと違っ
    # ているため。具体的には「小さい/小ちゃい」のような変種までいっ
    # しょくたにしているが、こういうのは別物として考えたいのだ。これは
    # 最終段階での分類に影響する。なお、代表表記が同じでも「言い難い」
    # に対してご丁寧に「イイガタイ」「イイニクイ」を与えているので注意。
    # これを「小さい/小ちゃい」と同類と考えるかは正直微妙なところだが、
    # 区別したい、という点では同じである。
    #
    # たとえば「言い難い」に関し、「イイガタイ」と「イイニクイ」の
    # 優劣はないのだが、これは jumandic での扱いに優劣がないことが
    # あるのと同じく、実際の日本語としても優劣がない場合もある。して、
    # これの「いずれかが」ipadic に既存だとする。この場合、「イイニクイ」
    # が既存の場合は「イイガタイ」が読み違いの変種である。この場合はいい
    # のだが、問題はともに既存でない場合である。優劣がないのだから、
    # 「代表表記以外」という言い方では片付けられない(し処理上も区別出
    # 来ない)。なので、「複数の読みがあるうちの一つである」ということが
    # 識別出来れば良いかなと。便利かどうかはともかく、これ以上のことは
    # 出来そうにないし。
    #
    # ちなみに、ほんとは「優劣はない」ことはなくて、ある場合もある。必ず
    # あるわけじゃないのが悩ましいが、少なくとも順位付け自体は(jumandic
    # のコストを信じるなら)出来る。というわけで、その順序付け実装は
    # 入れてはいるけれど、けれどもこの頑張りは実はデッドコード。偶然、
    # というには出来過ぎなので、おそらくまさに同じ考えに基くコスト調整
    # をやってる。それは_sortkey1 の並び順と同じになる調整。いやん…。
    # 「凄い」に関してはこんな具合:
    #
    #    凄い,1/1 スゴイ,凄い,1/2 スゴイ,凄い,-2597,...
    #    凄い,1/1 スゴイ,すごい,1/2 スゴイ,すごい,6735,...
    #    凄い,1/1 スゴイ,すっごい,2/2 スッゴイ,すっごい,10535,...
    #
    # 後者は、そもそも「jumandic の付与しているコストがあんましあてに
    # ならない」ことに起因している。というか「部分的にあてになる」こと
    # が、カユイというかなんというか。今度は「小さい」を例にすると、
    #
    #    小さい,1/1 チイサイ,小さい,1/3 チイサイ,小さい,5129,...
    #    小さい,1/1 チイサイ,ちいさい,1/3 チイサイ,ちいさい,10535,...
    #    小さい,1/1 チイサイ,ちっさい,2/3 チッサイ,ちっさい,10559,...
    #    小さい,1/1 チイサイ,ちっちゃい,3/3 チッチャイ,ちっちゃい,10535,...
    #    小さい,1/1 チイサイ,小ちゃい,3/3 チッチャイ,小ちゃい,4360,...
    #    小さい,1/1 チイサイ,小っさい,2/3 チッサイ,小っさい,4360,...
    #    小さい,1/1 チイサイ,小っちゃい,3/3 チッチャイ,小っちゃい,4360,...
    #
    # こんな具合。mecab-dict-train は当然全体でコストをアサインするの
    # だから、こういう個々に関しては無頓着なわけで(当たり前)、だから
    # 「オレは「小っさい」なんかいらんのだ、「小さい」だけオクレヨ」と
    # いうニーズにはそもそも応えてくれない。(「小っさい」の方が「小さい」
    # よりコストが小さい。衝突可能性から考えても「小っさい」の方がぶつ
    # かりにくかった、というトレーニング結果はわからんでもない。)
    # ゆえに、「小さい vs 小っさい」に関しては諦めるしかなくて、
    # 「小っさい vs ちっさい」の順序付けだけしようかと。当然これだけ
    # でも期待通りかといえばそんなことはなく、冒頭で書いた「因果応報」
    # では「んな綴りするかいな」というものが低コストになってるし、
    # 「いやいや、こっちの方がメジャーっしょ」が高めのコストになってる。
    # まぁ要するに、取捨選択のあんまし目安にもならんのだけれど、
    # 多少の傾向をつかむヒントくらいにはなるかなと。
    #
    # なお、処理上の問題として、本当は代表表記は「代表表記:言い難い/
    # いいにくい」という具合に漢字と読みを併記しているのでこれを使う
    # 手もあるのだが、いかんせん「入ってないのがいる」関係で、どっちが
    # 楽かわからん状態。ゆえに自力で読み解いてる。
    #
    ckeys = ("品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3")
    results0 = defaultdict(list)
    blddir = posixpath.join(args.builddir, "2nd")
    if not os.path.exists(blddir):
        os.makedirs(blddir)
    def _jdreader0(fn):
        for jdent in csv.DictReader(
            io.open(fn, encoding="utf-8"),
            fieldnames=fieldnames_all_juman_1st):
            yield from _jdreader_filter(jdent)()
    for fn in sorted(glob(posixpath.join(args.builddir, "1st", "*.csv"))):
        for jdent in _jdreader0(fn):
            for mvals in sorted(_omap.keys(), key=lambda x: (-len(x), x)):
                omkbase = tuple((jdent[k] for i, k in enumerate(ckeys) if i < len(mvals)))
                if omkbase == mvals:
                    results0[omkbase].append(jdent)
                    break

    # ---
    # 代表表記(読み)、原形(読み)の追加
    def _sortkey1(e):
        # 代表表記(読み)、原形(読み)を追加するのに必要なソート。
        return (
            (e["品詞"], e["品詞細分類1"], e["品詞細分類2"], e["品詞細分類3"], e["活用形1"]),
            e["代表表記"],
            (("0", "0") if e["活用形2"] == "基本形" else (e["活用形2"], e["-活用形2内部用-"])),
            (-len(e["読み"]), e["読み"]),
            e["代表表記"] != e["原形"],
            e["原形"],
            )
    results1 = defaultdict(list)
    for ck in results0.keys():
        entries = sorted(results0[ck], key=_sortkey1)
        outer = [  # 大枠で代表表記で括る。
            (g[0], list(g[1]))
            for g in itertools.groupby(
                entries, lambda x: (x["代表表記"]))]
        #
        for dh, oents in outer:
            inner = [  # 内側では活用形2で括る。
                (g[0], list(g[1]))
                for g in itertools.groupby(
                    oents, lambda x: (
                        x["活用形2"], x["-活用形2内部用-"]))]
            #
            # 基本形から代表表記の読みと原形の読みを。
            # ただしこれは読みそのものではなくて、
            # 変種の何番目かがわかるようにしておく。
            # (「1/2 イイニクイ」のように。)
            dhy, dhylstwc, ofylstwc = "", [], []
            for e in inner[0][1]:
                if dhy != e["読み"]:
                    dhy = e["読み"]
                cost = int(e["コスト"])
                dhylstwc.append((dhy, cost))
                ofylstwc.append((e["読み"], cost))
            #
            dhyvars = [x[1] for x in sorted([
                (min([x for x in g[1]]), g[0])
                for g in itertools.groupby(
                    sorted(dhylstwc), lambda x: x[0])])]
            ofyvars = [x[1] for x in sorted([
                (min([x for x in g[1]]), g[0])
                for g in itertools.groupby(
                    sorted(ofylstwc), lambda x: x[0])])]
            dhylst, ofylst, rnks = [], [], []
            for i in range(len(dhylstwc)):
                dhylst.append("{}/{} {}".format(
                        dhyvars.index(dhylstwc[i][0]) + 1,
                        len(dhyvars),
                        dhylstwc[i][0]))
                ofylst.append("{}/{} {}".format(
                        ofyvars.index(ofylstwc[i][0]) + 1,
                        len(ofyvars),
                        ofylstwc[i][0]))
                #
                # 同一「原形(読み)」内順位
                csof = sorted(
                    list(
                        set([c for y, c in ofylstwc if y == ofylstwc[i][0]])))
                rnks.append(
                    "{}/{}".format(
                        csof.index(ofylstwc[i][1]) + 1, len(csof)))
            # うめる、生める、産める、…、埋る、埋める
            for _, ents in inner:
                for i, e in enumerate(ents):
                    e["代表表記(読み)"] = dhylst[i]
                    e["原形(読み)"] = ofylst[i]
                    e["同一原形(読み)内順位"] = rnks[i]
                    results1[ck].append(e)

    # ---
    # 目的ブツの生成
    def _sortkey2(e):
        # 目的のソート。
        # ちょっとややこしいのだが、「小さい/小ちゃい」に関して「区別したい」
        # けれども、「ipadic 既存エントリの変種である」という情報は失いたく
        # ないわけだ。確かに「小ちゃい」なんてエントリは ipadic にはいないが、
        # 区別出来ない状態での新規エントリ扱いはイヤなわけだ。なのでここでの
        # ソートは代表表記(読み)、原形(読み)はトップレベルのソートキーには
        # しない。(「代表表記」だけでグルーピングするため。)
        return (
            (e["品詞"], e["品詞細分類1"], e["品詞細分類2"], e["品詞細分類3"], e["活用形1"]),
            e["代表表記"],
            (("0", "0") if e["活用形2"] == "基本形" else (e["活用形2"], e["-活用形2内部用-"])),
            e["代表表記"] != e["原形"],
            e["代表表記(読み)"],
            e["原形"],
            e["同一原形(読み)内順位"],
            e["原形(読み)"],
            e["表層形"],
            e["読み"],
            -len(e["出典等"]))
    for ck in results1.keys():
        entries = results1[ck]
        ofnbase, _ = _omap[ck]
        ofn = posixpath.join(blddir, ofnbase + ".csv")
        logging.info("dumping preprocessed output %r", ofn)
        with io.open(ofn, "wb") as fo:
            fo.write(("\n".join([
                            ",".join([ent.get(k, "") for k in fieldnames_all_juman_2nd])
                            for ent in sorted(entries, key=_sortkey2)
                            ]) + "\n").encode("utf-8"))


# ==================================================================
# 第三段階関連のもろもろ
# ==================================================================
def _jdreader(fn):
    yield from csv.DictReader(
        io.open(fn, encoding="utf-8"),
        fieldnames=fieldnames_all_juman_2nd)


class _yomi_cmp(object):
    # これは合致判定で使う。ipadic と読みだけが異なっている場合が
    # あり、これを許容できるなら許容するわけだが、いくつかのものは
    # 許容したくないわけだ。
    #
    # ひとつには、一方で読みをブランク(*)にしていたり、表層形を
    # そのまま入れているもの。このケースは全部記号や全角英数字のみ
    # エントリである。たとえば「ＧＨＱ」に対して、一方がブランクで
    # 一方が「ＧＨＱ］の場合、これを許容するのは無駄エントリ追加で
    # しかない。対して、ipadic 側で「ＧＨＱ］、jumandic 側で「ジー
    # エッチキュー」だったとすれば、これは少しはいいかもしれない。
    #
    # もうひとつはもちろん「表記ゆれ」の範疇に入るもの。これは
    # 完全なテストは非常に困難なのだが、簡単なものであれば検出出来る。
    # 「非常に困難」なのは、長音・拗音・促音に関して「パーティション/
    # パーテイション/パーテション/パーテエション/パーテーション」の
    # ようなケースがあるから。こうした長音・拗音・促音は消えたり
    # 化けたりするので一筋縄ではいかない。けれどもここはそんなに
    # 労力をかけるようなことでもないので、簡単な正規化でキャッチ
    # 出来るものだけにする。
    #
    # あとこれはもうどうしようもないのだが、ipadic には読みに関する
    # ミスは結構多くて、そうしたものはもうどうしようもない。(別のも
    # のからのコピペ後の修正もれと思われるものが多数ある。)
    #
    # ちなみに。「読みだけが違うエントリ」って、ほんとは MeCab 利用
    # においては、「どうでもいい人の方が多い」はずなんだよね。普通は
    # MeCab に期待するのは「トークンに分解」することであって、読みは
    # ほとんどオマケ。どうせコストが最も低いものがたった一つだけ
    # 選ばれ続けるんだから、選ばれない読み違いは永遠に眠ってる。
    # 私がこれを許容して取り込む理由は「オレさまのため」ではなくて、
    # どっちかつーとワタシとノリが違う人、つまり「システム辞書の
    # 方をゴニョゴニョしたい人」のため。そうした人は「いやいや、
    # こっちの読みが優先されないと困るっしょ」てことがしたいであろ。
    # そのためのネタ元としては有用だろうな、と思うわけだ。
    def __init__(self, hin, sur, rhs):
        self._hin = hin
        self._sur = sur

        # 取り込んで価値があるかどうか。表層形と読みが同じならば、「変種」
        # として取り込む価値はない。ちょっとややこしいのが「ＩＭＡＧＩＣＡ］。
        # これは無論「イマジカ」だが、jumandic ではそのまま「ＩＭＡＧＩＣＡ］、
        # これを _yomi_normが「アイエム…」と読み替えてしまうので、違いが出て
        # しまう。もちろん読み替えをしなければしないで違いが出る。この場合は
        # ipadic のほうに従う。(ほんとにこれは「ＩＭＡＧＩＣＡ］だけ。)
        self._valualble = _h2k(sur) != rhs and _yomi_norm(
            self._hin, self._sur, self._sur) != rhs

        #
        self._rhs = self._norm(rhs)

    def _norm(self, s):
        for f, t in (
            ("ー", ""),
            # ッはちょっと微妙。Prefix の「実」の「ジツ vs ジッ」。
            # けどまぁこれはいいでしょ。
            ("ッ", "ツ"),
            ("ァ", "ア"), ("ィ", "イ"), ("ゥ", "ウ"), ("ェ", "エ"), ("ォ", "オ"),
            ("ャ", "ヤ"), ("ュ", "ユ"), ("ョ", "ヨ")):
            ("ヅ", "ズ"),
            ("ヂ", "ジ"),
            s = s.replace(f, t)
        # 記号とか全角英数
        s = re.sub(r"^[^ァ-ンヴー]+$", "*", s)
        return s

    def __call__(self, lhs):
        if lhs == self._rhs or not self._valualble:
            return True
        if self._rhs == "*":
            return True
        lhs = self._norm(lhs)
        return lhs == self._rhs


def _rm_alnum(s):
    return re.sub(r"[A-Za-z0-9£]", "", unicodedata.normalize("NFKC", s))


def _norm_sur(yomi, s):
    # _kvar_simplifier による分類の本質に無関係な揺れの標準化。
    # 「・」有無は判定にはなんの関係もなく、また、「駒ケ岳」が
    # 「コマ「ガ」タケ」、「五ヶ瀬」は「ゴ「カ」セ」と読む、な
    # 措置が必要なの。ただし、「怖ケ付ク」なんてのもあるので注意。
    # jumandicでは下記正規表現でつかまるものが全て。
    s = re.sub(r"・", "", s)
    ga = ""
    if yomi.count("ガ") == 1:
        ga = "が"
    elif yomi.count("カ") == 1:
        ga = "か"
    if ga:
        s = re.sub(
            r"([^ぁ-んァ-ンヴー])[ケヶ]([岳丘岡峰宿浜瀬所谷里崎島根沢浦原])",
            r"\1{}\2".format(ga),
            s)
    return s


class _kvar_simplifier(object):
    # 「阜偏/小里偏/こざと偏/阜へん/小里へん」のような変種群で
    # 変種の区別をするのに使うやーつ。

    def __init__(self, dh, yomi):
        class surfs(object):
            def __init__(self, yomi):
                self._yomi_raw = yomi
                self._items = defaultdict(list)
                self.has_allkana = None
                self.has_kanji = None
                self.anyhira = False
                self.allhira = True

            def append(self, sur):
                self._items["raw"].append(sur)

                h = re.match(r"^[ぁ-んー]+$", sur)
                self.allhira = h and self.allhira
                self.anyhira = h or self.anyhira

                sur = _norm_sur(self._yomi_raw, sur)
                self._items["norm1"].append(sur)
                sur = _h2k(sur)
                self._items["norm2"].append(sur)

                self.has_allkana = re.match(
                    r"^[ァ-ンヴー]+$", sur) or self.has_allkana

                nln = _rm_alnum(sur)
                self.has_kanji = re.search(
                    r"[^ァ-ンヴー]", nln) or self.has_kanji

                self._items["rm_alnum"].append(nln)
                self._items["rm_kanji"].append(
                    re.sub(r"[^ァ-ンヴー]", "", sur))
                self._items["rm_kana"].append(
                    re.sub(r"[ァ-ンヴー]", "", sur))
                self._items["rm_alnumkana"].append(
                    re.sub(r"[ァ-ンヴー]", "", nln))

            def __getitem__(self, key):
                return self._items[key]

        self._yomi_raw = yomi

        # 代表表記がちとややこしい。jumandic エントリ自身が主張する
        # 代表表記そのものは当然必要なのだが、「小っさい」に対する
        # 代表表記「小さい」を避けたいがために、または、「代表表記
        # 読み」でグルーピングしているために update() に入ってこない
        # がために、の二つの理由により、「読み替え代表表記」が欲しい
        # の。
        self._dh = _norm_sur(self._yomi_raw, dh)
        self._dh_orig = self._dh

        self._yomi = tuple(enumerate(yomi))
        self._minidic = set()
        self._surfs = surfs(self._yomi_raw)
        self._classified = []
        self._has_allkana = False
        self._has_dkanji = ""

    def _update_dic(self, p):
        if not (p[0] and not re.match(r"^[ァ-ンヴー]+$", p[0])):
            return
        for kn, y in list(self._minidic):
            if y and p[1] and p[1][0] == y[0] and re.sub(r"[ァ-ンヴー]", "", p[0]) == re.sub(r"[ァ-ンヴー]", "", kn):
                if p[1] == y and len(p[0]) > len(kn) or \
                        p[0] == kn and len(p[1]) < len(y):
                    self._minidic.remove((kn, y))
                    self._minidic.add(p)
                break
        else:
            self._minidic.add(p)

    def _update(self, sur, yomi):
        def _tw(s1, s2):
            return list(itertools.takewhile(
                    lambda x: x[0] == x[1][1], zip(s1, s2)))
        if not self._yomi or sur == self._yomi_raw:
            return
        s, y = sur, yomi
        lm = len(_tw(reversed(s), reversed(y)))
        if lm:  # 「ヒッ切リ無イ」→「ヒッ切リ無」
            s, y = (s[:-lm], y[:-lm])
            self._update_dic((s, y))
        fm = len(_tw(s, y))  # 「ヒッ切リ無」→「ヒッ」「切リ無」
        self._update_dic((s[fm:], y[fm:]))

    def update(self, sur):
        self._surfs.append(sur)

    def _additional(self):
        # jumandic の機械生成(と思しき)生成が中途半端というか、数学的な
        # 全集合にはなっていなくて。やるなら全部やれよ、と文句のひとつも
        # 言いたくなる。たとえば「立看板」に対して「立てかん板」なんての
        # を律儀に投入しようとする癖に、「たて看板」がいない、とか。この
        # class はそもそもその「律儀な機械生成」を逆手に取るノリなので、
        # この不完全さが仇。ここではそれを(あくまでもベストエフォートで)
        # 補う。

        for t in tuple((p for p in list(self._minidic))):
            sur, yomi = t
            sur, yomi_raw = sur, "".join([c for _, c in yomi])
            if not re.match(r"^[^ァ-ンヴー]+[ァ-ンヴー]+[^ァ-ンヴー]+$", sur):
                continue
            # 以下やり過ぎると危険な一致をする可能性が高いため、非常に無難で
            # 穏健なことしか出来ない。

            # 「目ざまし時計」の場合の「ざまし」のように、仮名共通部分の出現
            # が一回きりのケースは、「目(メ)」+「ざまし」であると確定できる。
            kana = re.sub(r"[^ァ-ンヴー]", "", sur)
            if yomi_raw.count(kana) != 1:
                continue
            idx1 = sur.index(kana)
            idx2 = yomi_raw.index(kana)
            self._update(sur[:idx1], yomi[:idx2])
            self._update(sur[idx1:], yomi[idx2:])

        # その他はちょいと複雑。既出の断片を使ってさらに分解する。
        cont = True
        while cont:
            cont = False
            odlen = len(self._minidic)
            for t in list(self._minidic):
                sur, yomi = t
                for u in [u for u in list(self._minidic) if len(u[1]) < len(t[1])]:
                    if t[0].endswith(u[0]) and t[1][-len(u[1]):] == u[1]:
                        # 「荒れ模様」「模様」→「荒れもよう」
                        nent = list(t[0][:-len(u[0])])
                        nent += [c for _, c in u[1]]
                        nent = "".join(nent)
                        klen = len(list(itertools.takewhile(
                                    lambda x: re.match("[ァ-ンヴー]", x),
                                    reversed(nent))))
                        if nent[-klen:] == ("".join([c for _, c in yomi]))[-klen:]:
                            self._update(nent, yomi)
                    elif t[0].startswith(u[0]) and t[1][:len(u[1])] == u[1]:
                        # 「和気藹々」「和気」→「わき藹々」
                        nent = [c for _, c in u[1]]
                        nent += list(t[0][len(u[0]):])
                        nent = "".join(nent)
                        klen = len(list(itertools.takewhile(
                                    lambda x: re.match("[ァ-ンヴー]", x),
                                    nent)))
                        if nent[:klen] == ("".join([c for _, c in yomi]))[:klen]:
                            self._update(nent, yomi)
            cont = (len(self._minidic) > odlen)

    def _reconstruct(self):
        # _additional とかなり同じ正反対なことをする。つまりはやりたいことは、
        # たとえば「概ね/大旨」に対して「概ネ/オオ旨/大ムネ」ではなく
        # 分解しきったエントリを使って「概ネ/大旨」に再構成し直すことである。

        # まず、カナを含むエントリを除外。残るものが「最小単位」である。
        self._minidic = set([t for t in self._minidic
                             if not re.search(r"[ァ-ンヴー]", t[0])])

        # で、「概ね/大旨」の場合たとえば「オオム: 概」「オオ: 大」の
        # ようにバラされているので、ここから「オオムネ: 概ネ/大旨」に
        # 構成し直す。処理は結構ややこしいが、やりたいこと、やるべきことは
        # 至ってシンプルである。
        def _krepl(k):
            i, res = 0, ""
            while i < len(k):
                for kn, y in sorted(list(self._minidic)):
                    if len(y) == len(list(itertools.takewhile(
                                lambda x: x[0] == x[1], zip(k[i:], y)))):
                        res += kn
                        i += len(y)
                        break
                else:
                    res += k[i][1]
                    i += 1
            return res
        #
        for t in sorted(list(self._minidic)):
            first = t[1][0]  # (2, 'ブ')
            for i in range(first[0]):
                pre = self._yomi[i:first[0]]  # (0, 'ウ'), (1, 'ツ')
                k = pre + t[1]
                for t2 in [t2 for t2 in self._minidic if len(t[1]) < len(t2[1])]:
                    if k == t2[1] and t[0] not in t2[0]:
                        sur = _krepl(pre) + t[0]
                        self._minidic.add((sur, k))
        for t in sorted(list(self._minidic)):
            last = t[1][-1]  # (0, 'オ'), (1, 'オ') の (1, 'オ')
            for i in range(len(self._yomi) - last[0]):
                post = self._yomi[last[0] + 1:last[0] + i + 1]  # (2, 'ム')
                k = t[1] + post
                for t2 in [t2 for t2 in self._minidic if len(t[1]) < len(t2[1])]:
                    if k == t2[1] and t[0] not in t2[0]:
                        sur = t[0] + _krepl(post)
                        self._minidic.add((sur, k))

    def _omi_special(self):
        # 「御御/御神/お御/お神」までを「オミ」と読ませるところまでは
        # ともかくとして、「御」で「オミ」って、ほんとにあるか?
        # 幸いこのパターンは「オミクジ」のみ。本来は「オミオツケ」だの
        # 「オミグシ」だのいっぱいあるはずなんだけどね。なぜだか jumandic
        # ではこれしか思いつかなかったよーで。
        self._minidic = set()
        omi = ("御御", "御神", "オ御", "オ神")
        omi_kanjis = set()
        for o in omi:
            omi_kanjis |= set(list(o))
        for s in omi:
            self._minidic.add((s, self._yomi[:len("御御")]))
        for s in self._surfs["norm2"]:
            for o in omi:
                if s[:len(o)] == o and s[len(o):][0] not in omi_kanjis:
                    if not re.match(r"^[ァ-ンヴー]+$", s[len(o):]):
                        self._minidic.add((s[len(o):], self._yomi[len("御御"):]))

    def _check_dkanji(self):
        for fg in itertools.groupby(
            sorted([(y, kn) for kn, y in self._minidic]),
            lambda x: x[0]):
            fg1 = list(fg[1])
            if any([p for p in fg1 if not _rm_alnum(p[1])]):
                if self._surfs.has_allkana or len(fg1) > 1:
                    self._has_dkanji = "A"
            elif len(fg1) > 1:
                self._has_dkanji = "K"

    def _classify_prep(self):
        def _filtered_dic():
            return sorted(
                [p for p in self._minidic],
                key=lambda x: (
                    not re.search(r"[ァ-ンヴー]", x[0]) and not re.search(r".[々]", x[0]),
                    len(x[0]),
                    len(x[1]),
                    x[1][0],
                    x[0]))
        #
        repls = []
        for kn, yomi in _filtered_dic():
            repls.append((kn, "".join([y for _, y in yomi])))
        simp1 = list(self._surfs["norm2"])
        simp2 = list(self._surfs["rm_kanji"])
        simp3 = list(self._surfs["rm_kana"])
        for i in range(len(simp1)):
            for fr, to in repls:
                simp1[i] = simp1[i].replace(fr, "<{}>".format(to))
        dhidx = self._surfs["norm1"].index(self._dh)
        simp1_dh = simp1[dhidx]
        simp2_dh = simp2[dhidx]
        simp3_dh = simp3[dhidx]
        return dhidx, simp1_dh, simp1, simp2_dh, simp2, simp3_dh, simp3

    def _classify(self):
        # 変種の分類をする、てわけなんだけれども、これこそがワタシが
        # この「スクリプトらしきなにか」をやろうと思った最大の動機。
        # 「こんなんいらん」さえ捨てれれば、jumandic には結構役に立つ
        # エントリは多いわけなんだよね。悪貨が良貨を駆逐するつーかね。
        # たった2割の「なんだよこれ全然いらねーしトラブルメーカだよ」
        # のために、残り8割ももろとも捨てる「のには忍びない」って貧乏性
        # ね。
        #
        # 送り仮名、仮名化がとりわけ迷惑なやつ。例によって「ぜーんぶ
        # いらないっ」とはならないのが痛いが、ただ、こいつらの9割以上
        # ははっきりいってゴミ。仮名化の方がより大迷惑かというとそう
        # でもなくて、送り仮名違いの方が遥かに「厄介」。この「スクリ
        # プトとか…」の処理が送り仮名違いの扱いで苦慮したのと同じく、
        # MeCab 辞書に投入して利用する際にも同じトラブルを持ち込みうる。
        # たとえば「投げ打つ」に対する送り仮名違い「投うつ」はこれは、
        # 誰もが咄嗟に「トウウツ」と読むであろうし、「聞き苦」のそれの
        # 「聞苦」は、場合によっては「聴聞苦痛だ」から「聞く苦しい
        # だもんねーヴぁーかヴぁーか」と切り出しちゃうだろう。うん、
        # 例が良くなくてすまん。なんだかんだ「因がおう報」なんて、
        # 衝突可能性が低いからあんまし実害なかったりするんだけれど、
        # 送り仮名違いは実害が結構大きくなる、と思う。
        dhidx, simp1_dh, simp1, simp2_dh, simp2, simp3_dh, simp3 = self._classify_prep()
        self._classified = [{"gr1": [], "gr2": []} for _ in simp1]
        dh_iskana = re.match(r"^[ァ-ンヴー]+$", _h2k(self._dh))

        # 「従兄弟,いとこ,従兄,従妹,従姉,従姉妹,従弟」パターンのように、
        # ひらがなエントリ以外の全てが漢字でちゃんと異漢字のケースは、
        # 何か補足的に言う必要はないし、言っちゃダメ。全部が然るべき
        # 「欲しい」変種なのだから。
        tmp = [gb for gb in itertools.groupby(
                sorted(list(self._minidic), key=lambda p: p[1]),
                lambda p: p[1])]
        allkanji = len(tmp) == 1 and tmp[0][0] == self._yomi
        #
        leng = len(simp1)
        lhs = simp1_dh
        chf = lambda s: len(re.sub(r"[<>]", "", s))
        for i in range(leng):
            rhs = simp1[i]
            if not self._surfs["rm_alnum"][i]:
                continue
            of = self._surfs["norm1"][i]
            if self._surfs["rm_alnumkana"][dhidx] and \
                    re.match(r"^[ぁ-んー]+$", of):
                # 代表表記が漢字等交じりのオールひらがなバージョン。
                self._classified[i]["gr2"].append("全平仮名化")
                continue
            if not self._surfs.has_kanji:
                # 仮名のみで構成される語。
                if self._surfs.anyhira and not self._surfs.allhira:
                    # かなのみで構成される語のうち、「カタカナ語」でない
                    # もの。さすがに「おーすとらりあ」なんてのはないので
                    # そこは安心。たとえばこれに該当するのは「ださい/ダサい」
                    # とかそんなやつ。これについては、優劣は付けない。つまり、
                    # 「ウヨウヨ/うようよ」に関し、「うようよ」が「ウヨウヨ」
                    # のひらがな化である、かどうかなんか…どうでもいいでしょ?
                    # 変種がある、その変種のうちどっちだ、とわかればいいのだ。
                    if re.match(r"^[ぁ-んー]+$", of):
                        self._classified[i]["gr1"].append("ひらがな")
                    else:
                        self._classified[i]["gr1"].append("カタカナ")
                elif self._dh_orig != of:
                    # かなのみで構成される語でなおかつ代表表記と違う、となれば、
                    # 長促拗音等の変種。有用な変種も割とある一方で、「アイス
                    # ホッケー」に対する「アイスホッケ」のようなかなり意味わか
                    # らんのも大量に。こういうのは未知語でええやん、と思うんだ
                    # けどねワタシは。
                    # 非常に微妙なのもある。「そこら」に対する「そこいら」。
                    # こういう変種は確かに有用と思う。ゆえ、やはり「このタイプ
                    # の変種は全部いらない」とはなかなか出来ない。弱るねぇ。
                    self._classified[i]["gr1"].append("長促拗音等")
                continue
            if dh_iskana and self._surfs["rm_alnumkana"][i]:
                # 代表表記がカナの場合の漢字バージョン。こう
                # いうのは概ね漢字バージョンの方がヘンチクリン。
                self._classified[i]["gr2"].append("漢字表記")
            if chf(lhs) > chf(rhs):
                s = self._surfs["norm2"][i]
                if re.match(r"^[^ァ-ンヴー]+$", _h2k(self._dh)) and re.match(r"^[^ァ-ンヴー]+$", s) and \
                        len(self._dh) > len(s) and _h2k(self._dh).startswith(s):
                    # 「大晦日/大晦」「柚子/柚」「目論見/目論」などのように、
                    # いくつか「異漢字変種」として扱う必要があるものがキャッチ
                    # 出来ていないのでここで。
                    self._has_dkanji = "K"
                elif not allkanji:
                    self._classified[i]["gr2"].append("送り仮名")
            if lhs != rhs and simp2[i] and len(simp2_dh) < len(simp2[i]) + (chf(lhs) - chf(rhs)):
                self._classified[i]["gr2"].append("仮名化")

    def end_update(self):
        if not self._yomi:
            return
        #
        if self._dh not in self._surfs["norm1"]:
            # 「有り丈」の代表表記が「有りっ丈」、「小ちゃい」「小っさい」
            # の代表表記が「小さい」となってる関係で、代表表記そのものが
            # 入ってこないことがある。「小ちゃい」と「小さい」を同列に扱う
            # 必要があるのかないのかだが、これは ipadic に「小さい」が既存
            # で「小っさい」が既存でない場合に「何も言わずに新規扱いするの
            # がイヤ」という理由で、「同列に扱いたい」。けれどもこれこそが
            # _kvar_simplifier にとって敵。弱るねぇ。「便覧」に対する
            # 「べんらん/びんらん」パターンの場合は、jumandic エントリ自身
            # が主張する代表表記をそのまま使えば良いのだけれど、正直いって
            # 「有り丈」の「有りっ丈」、「小ちゃい」「小っさい」の「小さい」
            # は邪魔くさい。拗促音を含む場合は特別扱いしとく?
            tail = len(list(itertools.takewhile(
                        lambda x: x[0] == x[1],
                        zip(reversed(_h2k(self._dh)),
                            reversed(self._yomi_raw)))))
            def _yss(s):
                return re.search(r"[ァィゥェォッャュョ]", s)
            def _rm_kana(s):
                return re.sub(r"[ァ-ンヴー]", "", _h2k(s))
            if _rm_kana(self._dh) == self._dh or \
                    (tail and _rm_kana(self._dh[:-tail]) == self._dh[:-tail]) or \
                    _yss(_h2k(self._dh)) == _yss(self._yomi_raw) or \
                    (tail and _yss(_h2k(self._dh[:-tail])) == _yss(self._yomi_raw[:-tail])):
                self._surfs.append(self._dh)
            elif re.match(r"^[ァ-ンヴー]+$", _h2k(self._dh)):
                self._dh = [
                    s for s in self._surfs["norm1"]
                    if re.match(r"^[ァ-ンヴー]+$", _h2k(s))
                    ][0]
            else:
                # 「秘蔵っ子」に対する「秘蔵=ひぞう」つまり:
                #     「dh: 秘蔵っ子/sur: ひぞうっ子」
                # という対応がないので、以下は大抵は漢字少な目のものを
                # 引っかけてしまう。けれどもだからといって
                # _yss(self._dh) == _yss(self._yomi_raw) 時と同じものを
                # 採用すると今度は「小っさい」の代表表記「小さい」を
                # 採るハメになり、これはイヤだ。(送り仮名判定等々が
                # 全部ぶっ壊れる。)
                # ということなんだけれども、「秘蔵っ子」に関しては、
                # 「「ひぞうっ子」よりも「ひぞうっこ」がイヤ」てことで
                # 妥協しとこう、てことだわ。
                self._dh = list(
                    sorted(
                        self._surfs["norm1"],
                        key=lambda x: (
                            -len(re.sub(r"[ァ-ンヴー]", "", _h2k(x))),
                            -len(re.sub(r"[^ァ-ンヴー]", "", _h2k(x))))))[0]
        #
        for sur in sorted(
            self._surfs["norm2"],
            key=lambda s: (
                -len(re.sub("[ァ-ンヴー]", "", s)),
                 -len(re.sub("[^ァ-ンヴー]", "", s)),
                 s,
                 )):
            self._update(sur, self._yomi)
        if any(["御御" in s for s in self._surfs["norm2"]]):
            # これはもうかなりどうしようもないパターン。中国四千年の
            # 特殊処理。
            self._omi_special()
        else:
            self._additional()
            self._reconstruct()
        #
        self._check_dkanji()
        self._classify()

    def has_dkanji(self):
        return self._has_dkanji

    def classified(self, sur):
        c = self._classified
        idx = self._surfs["raw"].index(sur)
        return c[idx]["gr1"], c[idx]["gr2"]


def _yomi_diffvar(sur, y1, y2):
    # _yomi_cmp と似ているが目的が少しだけ違う。ipadic と合致している
    # ときに、たとえば「罪深い」に対して「ツミフカイ」と読むか「ツミ
    # ブカイ」と読むか、などのバリエーションがあるわけだ。この差異を
    # 「意義深い」と思うならいいのだが、半分以上の MeCab 利用にとって
    # は、これはまったくもってどーでもいい。とりわけ「分かち書きさえ
    # できればええんや」タイプのユーザにとってはこれは邪魔でしかない。
    # ので、「読みの違いの価値」とでも言うんかな、これのちょっとした
    # 色分けをしたい、ってわけだ。こういう「わかっとるわい」てのを
    # 軽んじたいわけさね。
    #
    # ちょっと注意して欲しいのは、ワタシとしては、そもそもが「読みの
    # 違いなんぞどーでもいい」と思ってるってこと。ゆえに「読み違い
    # バージョン」を大分類レベルで追い出してる。それでもここでこと
    # さらに細分類しているのは、「ワタシよりも貧乏性」な人向けの配慮
    # ね。「欲しいのがあるかもしれん」て思う人が少しでも「本当に自分
    # が欲しいもの」を区別出来るように、ってわけだ。_yomi_diffvar で
    # 細分類されたものは、基本的には「本質的でない読み違い」であり、
    # ゆえに、ここで何かを「言ってないもの」が、おそらくあなたが「欲
    # しがるかもしれないもの」である。
    #
    # あと何度も繰り返しになるけれど、「ipadic vs jumandic」という
    # 対立構図に関してもワタシは何も言わない。はっきりいって「jumandic
    # の方が明らかに正解」というものもかなり多い。要するに「その正誤
    # そのものに、あなたの利用にとっての価値があるかないか」の話で
    # ある。読み上げ用途を考えているなら、当然「正しい読みが欲しい」
    # だろう。MeCab のユースケース次第、ってことね。(ワタシの場合は
    # 「分かち書きさえできれば」ではないんだけれど、「形態要素に
    # 狙い通りに分割されればいい」という使い方なので、だから読みは
    # 全然使わない。一番極端なケースだと「句読点かどうかだけわかれば
    # いい」という使い方をすることもある。この場合「・(テン)」なんて
    # ほんっとどーでもいい。)

    if sur == y2:
        return "記号の読み"

    if len(y1) > len(y2):
        y1, y2 = y2, y1
    
    if len(y1) != len(y2):
        if len(y2) - len(y1) == 1:
            if y2[:-1] == y1 and y2[-1] in ("ウ", "ー"):
                # 桜ん坊に対する「サクランボ/サクランボウ」など。
                return "ウー"
            elif y2.replace("ン", "") == y1:
                # 「ミヤスドコロ」と「ミヤスンドコロ」とか。
                return "ン"
            elif sur[0] == "御" and y2[0] == "オ":
                return "御オ"
            elif y2.replace("ッポン", "ホン") == y1:
                return "本"
            elif "太" in sur and y2.replace("ダイ", "ダ") == y1:
                return "太"

        # かなり別物。「取り得」を「トリエ」と読むか「トリドク」と
        # 読むか、てことだが、「取り得」が「取り得る」の活用の一部
        # でない場合に「トリエ」と読むことはなかろう。jumandic は
        # 「よくある読み間違い/書き間違い」にも異常に寛容で、なので
        # 「取り柄」の(誤字)変種として「取り得」を入れとるわけである。
        return ""

    def _d2s(c):
        d = "ガギグゲゴザジズゼゾダヂヅデドバビブベボパピプペポ"
        s = "カキクケコサシスセソタチツテトハヒフヘホハヒフヘホ"
        di = d.find(c)
        if di >= 0:
            return s[di]
        return c

    typ = set()
    for i in range(len(y1)):
        if y1[i] == y2[i]:
            continue
        so1 = tuple(sorted((y1[i], y2[i])))
        so2 = tuple(sorted((y1[i:i + 2], y2[i:i + 2])))
        if _d2s(y1[i]) == _d2s(y2[i]):
            typ.add("濁音")
        elif so1 == ("ズ", "ヅ"):
            typ.add("ズヅ")
        elif so1 == ("ス", "ツ"):
            # 「シリスボミ/シリツボミ」
            typ.add("スツ")
        elif so1 == ("ジ", "ヂ"):
            typ.add("ジヂ")
        elif so1 == ("ウ", "オ"):
            typ.add("ウオ")
        elif so1 == ("ウ", "ー") or so1 == ("オ", "ー"):
            typ.add("長音")
        elif so1 == ("ブ", "ム"):
            typ.add("ブム")
        elif so1 == ("オ", "ホ"):
            typ.add("オホ")
        elif so1 == ("ミ", "ン") or so1 == ("ム", "ン"):
            # 神無月
            typ.add("ン")
        elif i < len(y1) - 1 and y1[i] == "ッ" or y2[i] == "ッ":
            typ.add("促音")
        elif so2 == ("ジョ", "ニョ"): 
            typ.add("ジョニョ")
        else:
            typ.add("×")
    if "×" in typ or len(typ) > 1:
        # 二種類以上の共存は仮定しないし、実在もしてない。
        return ""
    if len(typ):
        return list(typ)[0]
    return ""


class _accumulator(object):
    _dkeys = (  # ipadic との合致検出に使うキー
        "表層形",
        "品詞",
        "品詞細分類1",
        "品詞細分類2",
        "品詞細分類3",
        "活用形1",
        "活用形2",
        "読み",
        )

    def __init__(self, fn, fromsysdic):
        self._fromsysdic = fromsysdic
        logging.info("processing preprocessed input %r", fn)
        self._reader = _jdreader(fn)

    def _detect_fromsysdic(self, jdent):
        sur = jdent["表層形"]
        if sur not in self._fromsysdic:
            return {}, []
        of = jdent["原形"]
        ipaents = self._fromsysdic[sur]
        ipaents = list(set([tuple(((k, ie[k]) for k in self._dkeys)) for ie in ipaents]))
        ipaents = [[v[1] for v in ie] for ie in ipaents]
        matched, conflicts = [], []
        for lhs in sorted(ipaents):
            rhs = [jdent[k] for k in self._dkeys]
            match_rank = -1  # 負でない場合、小さいほど優先度が高い合致、とする。
            if lhs == rhs:
                match_rank = 0
            elif rhs[1] == "記号":
                # 記号については ipadic と表層形が同じなら同じとみなす。
                # 違ったものとして登録しようとしてしまうと問題を起こす
                # ことがある。「句点」「括弧開・閉」だけは特別扱いされる
                # (BOS/EOS)ので。(変えるのも考え方で、事実私個人の
                # ユーザ辞書では「!」「?」をあえて「句点」として上書き
                # して使っている。つまりここで ipadic のを上書きして
                # しまうと、私の「オレ流句点」がもみ消されてしまいうる。)
                match_rank = 0
            elif lhs[:-1] == rhs[:-1]:
                # 読み以外一致
                match_rank = 3

            # ---
            # 以下二つの判定はちょっと込み入っている。基本的には、
            # 活用の異なる変種自体はほとんど許される。ipadic でも
            # 五段・ラ行と一段の両方がいるエントリとか結構多い。
            # これは無論自動詞・他動詞の違いであったり、同じ漢字・
            # 送り仮名でも読みが違うと活用が変わったり、あるいは
            # 古文専用の活用もある。けれども「特殊なものは特殊」
            # として、一部この自由が効かないものがある。それは、
            # rewrite.def で読み替え後の右辺となるもの、つまり、
            # たとえば「(ない|無い|いい|らしい)→無い」などで、
            # この対象になるものの多くが、個々に専用の left-id、
            # right-id が与えられている。実際に jumandic から
            # 持ってこようとして問題を起こすのは形容詞の「いい」
            # 「らしい」、動詞の「ゆく」の3つで、どうやら勝手な
            # 活用は出来ないらしい(mecab-dict-index に拒絶される)。
            # なので、これに該当するものは、ひとまず「別の活用
            # なんてさせないもんねー、ばーべろべろー」としとく。
            # ほんとはまじめに rewrite.def を読んで処理するのが
            # 理想だけれど、そこまでのことを必要としている
            # 「スクリプトを。。。」ではないので、ここはインチキに。
            elif rhs[1] == "形容詞" and lhs[1] in ("形容詞", "助動詞") and \
                    re.sub(r"^(ない|無い|いい|らしい)$", "無い", of) == "無い":
                logging.debug("\n%r\n%r\n", rhs, lhs)
                match_rank = 0
            elif rhs[1:3] == lhs[1:3] and rhs[-3] != lhs[-3] and of == "ゆく":
                logging.debug("\n%r\n%r\n", rhs, lhs)
                match_rank = 0
            # ---

            elif lhs[6] == rhs[6] and \
                    rhs[1] == "動詞" and rhs[2] in ("接尾", "非自立",) and lhs[1] in ("助動詞",):
                # jumandic 自身が接尾だと思っているもの、または私の読み替えで
                # これになったものが、助動詞に既存の場合がある。
                # どっちもいていいじゃないかと思うかもしれないが、これがなかなか
                # そうは言えなくて。コストのアサイン次第ではあるのだけれど、
                # 結びつきの強さが変化してしまうことによって、これの追加により
                # 時として悪影響が出てしまうのだ。ありがちなのは、ひらがなの
                # 多いワードをやたらに千切りたがってしまうようになる、とか。
                # なお、これに該当するのは「-ある」のみである。
                match_rank = 2

            elif (lhs[1:3], rhs[1:3]) == (["形容詞", "非自立"], ["形容詞", "接尾"]):
                # jumandic 側での区別が出来ないので一律「接尾」としたが、
                # ipadic に「非自立」がいれば、これは同一視する。
                match_rank = 1

            elif rhs[1] == "副詞":
                if (
                    (lhs[2], rhs[2]) == ("助詞類接続", "一般")):
                    # Demonstrative からの「副詞,一般」が ipadic の「助詞類
                    # 接続」ならば同じとみなす。
                    # 「細々」などでちょいとややこしい。ipadic では、
                    # 「ほそぼそ」(一般)、「こまごま」(助詞類接続)と
                    # 役割に応じて読みを変えている。対して jumandic は
                    # もともと細分類していないので、読み違いバージョンが
                    # あれば両方「一般」として入ってくる。ゆえに、
                    # 「細々(こまごま)」は工夫しないと「細々(ほそぼそ)」
                    # (一般)の読み違い変種になってしまう。ゆえに、
                    # 「読みだけが違う」よりもこちらの方を優先する。
                    match_rank = 2
                elif (
                    (lhs[1:3], rhs[2]) == (["副詞", "一般"], "助詞類接続")):
                    match_rank = 2
                    # 上の逆パターン。ない、はず。今のところそういう
                    # 読み替えはしてない。
                elif lhs[1] == "接続詞" and rhs[2] == "一般":
                    # 「あるいは」はさすがに接続詞だろう、と。
                    match_rank = 2
                elif lhs[1:3] == ["名詞", "副詞可能"]:
                    # ipadic に「名詞,副詞可能」がいれば同じとみなす。
                    # (「文字通り」など。)
                    match_rank = 4
                elif lhs[1:3] == ["名詞", "代名詞"]:
                    # ipadic に「名詞,代名詞」がいれば同じとみなす。
                    # (「あちこち」など。)
                    match_rank = 4
                elif lhs[1:3] == ["名詞", "形容動詞語幹"]:
                    # ipadic に「名詞,形容動詞語幹」がいれば同じとみな…していいと
                    # 思う、たぶん。
                    # (「いかが」など。)
                    match_rank = 4
                elif lhs[1:3] == ["名詞", "サ変接続"]:
                    # ipadic に「名詞,サ変接続」がいれば同じとみな…していいのかな。
                    # (「冷え冷え」「惚れ惚れ」など。)
                    # どちらが正しいとかじゃなくて、どっちかあればいいと思う。
                    match_rank = 4

            elif rhs[3] == "地域":
                if lhs[3] == rhs[3]:
                    if (
                        lhs[:4] == rhs[:4] and lhs[4] in ("国", "一般") and rhs[4] in ("国", "一般")) or (
                        lhs[:3] == rhs[:3] and lhs[3] == "一般" and rhs[3] == "地域"):
                        match_rank = 4
                elif lhs[2:4] == ["固有名詞", "一般"]:
                    # たとえば「洞爺湖」が ipadic には「固有名詞,一般」にいて、
                    # jumandic では地域扱いになっている。どちらが正しいという
                    # ことではなくて。
                    match_rank = 4

            elif lhs[1:4] == rhs[1:4] and lhs[3] == "人名":
                if rhs[4] == "一般":
                    # いまのところ外国人名のみ。この場合に ipadic に「姓」「名」区別
                    # された状態で登録されているなら jumandic にあるものと同じと
                    # みなす。(個々にひとつずつやるなら別だが、一括で考える場合は
                    # こうせざるを得ない。)
                    match_rank = 4
            elif (lhs[1:3], rhs[2]) == (["名詞", "一般"], "固有名詞"):
                # 「薬師如来」などは ipadic で「名詞,一般」、jumandic では固有名詞。
                match_rank = 4
            elif (lhs[2:4], rhs[2:4]) == (["固有名詞", "一般"], ["固有名詞", "組織"]):
                # どんな固有名詞なのかの区別を ipadic でしてないというだけ。
                match_rank = 4
            elif (lhs[2:4], rhs[2:4]) == (["固有名詞", "一般"], ["固有名詞", "地域"]):
                # 「洞爺湖」と逆パターン。一応処理上分岐は書いてるが、
                # このパターンはない。
                logging.info("%r, %r", lhs, rhs)
                match_rank = 4
            elif (lhs[2:4], rhs[2:4]) == (["固有名詞", "地域"], ["固有名詞", "一般"]):
                # 「八ツ橋」などは追加として許容したいが、「スラブ」などは
                # 「地域」のほうにいれば十分だと思う。
                match_rank = 4 if re.match(r"^[ァ-ンヴー・]+$", sur) is not None else 10
            #
            elif (lhs[2:4], rhs[2:4]) == (["接尾", "形容動詞語幹"], ["接尾", "一般"]):
                # jumandic 分類で「形容詞性名詞接尾辞」になっているものを
                # 「一般」にしたものに、ipadic では「形容動詞語幹」に分類されている
                # ものがある。(「的」など。)
                match_rank = 4
            elif (lhs[2:4], rhs[2:4]) == (["接尾", "特殊"], ["接尾", "形容動詞語幹"]):
                # jumandic 分類で「名詞性述語接尾辞」になっているものを一律
                # 「形容動詞語幹」にしたが、ipadic では「特殊」に分類されている
                # ことがある。(「っぷり」など。)
                match_rank = 4
            elif (lhs[2], rhs[2:4]) == ("接尾", ["接尾", "一般"]):
                if lhs[3] in ("副詞可能", "サ変接続", "特殊", "人名", "地域"):
                    # ipadic と jumandic のどちらの分類が「細かい」かは各々で、
                    # これらに関しては ipadic が細かい。 jumandic は人名にくっつく
                    # 接尾なのかどうか、副詞可能かどうか…を区別していない。ゆえに、
                    # ipadic にこれら接尾として存在していれば、同じものであるとみなす。
                    # 「地域」だけはちょっと特殊で、「郡」「区」「町」だけ
                    # なぜか jumandic で「地域接尾」として「固有名詞」ファイル
                    # にぶっ込まれてる。
                    match_rank = 4
            elif (lhs[1:3], rhs[2]) == (["名詞", "非自立"], "接尾"):
                # おそらく「名詞,非自立」は接尾とほぼ同じ役割だと
                # 思う。「～以外」や「～以上」などがこれにあたる。であれば、
                # jumandic での「接尾,一般」をあえて登録する必要があるかと
                # いえば、たぶんない…、と思うんだけれど、これは色々解析
                # させてみないとなんともいえないかも。
                if rhs[3] == "一般" or lhs[3] == "一般" or lhs[3] == rhs[3]:
                    match_rank = 5

            #
            elif (lhs[1] == "名詞" and lhs[2] in (
                    "代名詞", "サ変接続", "副詞可能", "形容動詞語幹",) and \
                      rhs[1:3] == ["名詞", "一般"]):
                # ContentW で「普通名詞」にされてるものでかなりの数が
                # ipadic で「代名詞」や「サ変接続」「副詞可能」などと
                # して既存。これは同じものとみなす。
                match_rank = 4

            #
            elif (lhs[1] == "名詞" and lhs[2] in ("ナイ形容詞語幹",) and \
                      rhs[1:3] in (["名詞", "サ変接続"], ["名詞", "一般"])):
                # ContentW で「一般」「サ変接続」にされてるもので「ナイ形容詞語幹」
                # として既存のものがある。これは同じものとみなす。両方いた
                # 方がいい可能性は…あるかもしれないけれど。たとえば
                # 「申し訳」は普通は「申し訳ない」だが、「申し訳する」で
                # 新撰組に狩られるわけでもなし。けどまぁいらんでしょう。
                match_rank = 4

            #
            elif (lhs[1] == "名詞" and lhs[2] in ("サ変接続",) and \
                      rhs[1:3] == ["名詞", "副詞可能"]):
                # ContentW で「副詞可能」にされてるもので「サ変接続」
                # として既存のものがある。これは同じものとみなす。
                # というかこれに該当するものはどれも日本語としてピンと
                # こないものばかりなので、なんか考えたくない。
                # (具体的には「もっか」とかの、なぜか全部ひらがなのみの
                # エントリ。)
                match_rank = 4

            elif (lhs[1:3], rhs[1:3]) == (["名詞", "一般"], ["名詞", "サ変接続"]):
                # たとえば「人見知り」は「サ変接続」にしておくほうが
                # 良いのは確かだが、「一般にしかいない」ことが支障が
                # あるかといえば…そうでもない気がする。結局「する」
                # が続けば「人見知り(名詞,一般) + する(動詞,自立)」
                # と分解されるのであろうし。
                # はっきりいってこれは非常に数が多いので、これを入れ
                # ちゃうとキリがないから、というのが本音。
                match_rank = 4

            elif (lhs[1:3], rhs[1:3]) == (["名詞", "一般"], ["名詞", "副詞可能"]):
                # (["名詞", "一般"], ["名詞", "サ変接続"])とほぼ同じ
                # パターン。「メーデー」などは「副詞可能?」と思うけれど、
                # 「末期」などは、「ふむふむ、確かにそうだね」と。
                # けれどもこれも数が多いので、まぁ一般にいれば十分
                # であろうと。
                match_rank = 4

            elif (lhs[1:3], rhs[1:3]) == (["名詞", "接尾"], ["名詞", "一般"]):
                # これは結構微妙なやつで、「まがい」など自明なものを
                # 除くと、ほんとに一般名詞として登録してあったほうが
                # よいのかもしれないものもある。「つき」などがそう。
                # jumandic は同じ意味のひらがなバージョンを積極的に
                # 入れるポリシーのようで、だからひょっとしたら「月」
                # のひらがな版のつもりなのかもしれない、など。
                # ただ、ざっくり眺めた感じだと、 ipadic にあるほう
                # だけで十分に思えた。ので、同じものとみなしておく。
                match_rank = 5

            elif (lhs[1:3], rhs[1:3]) == (["名詞", "代名詞"], ["名詞", "副詞可能"]):
                # 「時相名詞」を「副詞可能」に読み替えている関係で、
                # まさに「いつ」という代名詞まで副詞可能にしてしまって
                # いる。これは同じものとする。
                match_rank = 4

            elif (lhs[1:4], rhs[1:3]) == (["名詞", "接尾", "サ変接続"], ["名詞", "サ変接続"]):
                # 「視」。これは同じとみて良いだろう。
                match_rank = 6

            elif rhs[2] == "副詞可能":
                # Noun.hukusi からの「名詞,副詞可能」が ipadic の「名詞,
                # 非自立」にいればこれと同じとみなす。非自立の種類は
                # 問わない。
                # Noun.hukusi からの「名詞,副詞可能」が ipadic の「名詞,
                # 接尾」にいればこれと同じとみなす。接尾の種類は問わない。
                match_rank = 7 if (lhs[1:3] in (["名詞", "非自立"], ["名詞", "接尾"])) else -1
                # Noun.hukusi だけをみていたうちは『jumandic から「名詞,副詞
                # 可能」としたものが ipadic に「名詞,一般」「名詞,サ変接続」
                # で存在しているものがいくつかあるが、副詞可能としてもあった方
                # が良いだろうと思うので、重複とはみなさない。』と考えていた
                # が、ContentW での量をみて考え直した。

            #
            if match_rank >= 0:
                matched.append((match_rank, lhs))
            elif lhs[1] == rhs[1]:
                conflicts.append(lhs)
        return list(sorted(matched))[0][1] if len(matched) else [], conflicts

    def _accept(self, jdent, matched, conflicts):
        rhs = [jdent[k] for k in self._dkeys]
        ycmp = _yomi_cmp(jdent["品詞"], jdent["表層形"], rhs[-1])
        if matched:
            if matched[:-1] == rhs[:-1] and not ycmp(matched[-1]):
                # 読み以外一致する場合。
                return True
            logging.debug(
                "entry %r is already in ipadic as %r",
                "-".join([jdent[k] for k in self._dkeys if jdent[k] != "*"]),
                "-".join([m for m in matched if m != "*"]))
            return False
        return True

    def _basicform_info(self, gbkk2):
        # 「基本形」の英訳は basic form でいいんだろうか…。
        # ともあれ、ここでは活用語(用言)の場合は基本形、
        # 活用しない語(体言)についてはそれそのものに関しての
        # 情報収集をする。
        bfi = {
            "vars": [],  # 個々の変種について。
            }
        kvss = {}  # 読み: _kvar_simplifier
        for ent in gbkk2[0][1]:  # 基本形だけ調べる
            kvsskey = ent["原形(読み)"].split(" ")[1]
            if kvsskey not in kvss:
                kvss[kvsskey] = _kvar_simplifier(
                    ent["代表表記"], kvsskey)
            kvss[kvsskey].update(ent["原形"])
            #
            matched, conflicts = self._detect_fromsysdic(ent)
            acc = self._accept(ent, matched, conflicts)
            yomivar, varvar = False, False
            if matched:
                # 読みの違い。シンプルに読みが違うだけの場合が
                # ほとんどだが、jumandic の異常な送り仮名違いへ
                # の寛容さのせいで、実は送り仮名違いでこれに該当
                # するのがある。(「降す(クダス/オロス)」など。)
                # これの話がそう単純ではないのは、当然意味が違う
                # から読みが違う、という、日本語そのものが持つ
                # 本来の変種がまずあり、jumandic がそう考えるように
                # 送り仮名の違い(つまり送り仮名と読みが違うという
                # ことは普通意味も違う)もあり、なおかつ「表記ゆれ」
                # (わかりやすいのでは「ヅ」と「ズ」)があり、そして
                # 極めつけは「ミスがある」。この読み違いが引っかかる
                # のを詳しく観察していると、ipadic の誤りが結構
                # 見つかる。
                yomivar = ent["読み"] != matched[-1]
            # 活用する語の場合、品詞・細分類1が同じものがあれば。
            # 「意味が違うので読みも違い、活用も違う」というもの
            # がもちろん存在する。yomivar とセットでみることで、
            # 「活用のみ違う/読みと活用が違う」の区別が出来る。
            if conflicts and ent["活用形2"] != "*":
                rhs = [ent[k] for k in self._dkeys]
                varvar = any(((c[1:3] == rhs[1:3]) for c in conflicts))
            #
            bfi["vars"].append({
                    "ma": matched,
                    "ac": acc,
                    "hi": re.match(r"^[ぁ-んー]+$", ent["表層形"]) is not None,
                    "yv": yomivar,
                    "vv": varvar,
                    })
        for kvsskey in kvss.keys():
            kvss[kvsskey].end_update()

        # 変種個々ではなく、「基本形」全体に関する情報を補足。
        # また、体言に関しては、ipadic バージョンの品詞への
        # 読み替えをするので、その情報。
        # ややこしいが「変種たち」各々は「合致」していたり「合致
        # ではない」とされていたりして、原則「合致しているので
        # 拒否」なので「一部合致バージョンがあるうちの、拒否され
        # なかったものの扱い」の話なのだ。「変種扱い」なのだから、
        # 品詞等属性は一致してないとおかしい、てこと。
        sum = {
            # いずれかは ipadic との重複ではないと許容。
            "anyacc": any([b["ac"] for b in bfi["vars"]]),

            # いずれかは ipadic との重複。
            "anymatch": any([b["ma"] for b in bfi["vars"]]),

            # オールひらがな以外との合致があるかどうか。
            # 「綴り違い」とみなすかどうかは、「オールかな」
            # でないバージョンがある場合は「オールかな」
            # 以外との合致があるかどうか。「オールかな」
            # でしか合致しない場合は、素直に新規エントリにしな
            # ければならない。(たとえば ipadic に「-み (接尾)」が
            # いるとして、 jumandic が「-み」と「-味」を登録
            # しようとしている場合、ipadic にひらがな版がいようが
            # 新規エントリとして良いであろう。不愉快な一致の例が
            # 思いつかないが、要するに、読みだけでは合致かどうかの
            # 判断材料として少な過ぎるということ。)
            "matched_nonhira": any(
                [not b["hi"] for b in bfi["vars"] if b["ma"]]),
            "allhira": all([b["hi"] for b in bfi["vars"]]),

            #
            "kvss": kvss,
            }
        bfi.update(sum)
        #
        newattrs_keys = (
            "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3")
        newattrs = []
        if bfi["anyacc"] and bfi["anymatch"]:
            cands = set([
                tuple(b["ma"][1:-1]) for b in bfi["vars"]
                if b["ma"] and (not b["hi"] or not bfi["matched_nonhira"])])
            if len(list(cands)) == 1:
                newattrs = list(cands)[0]
            elif len(set([c[0] for c in cands])) == 1 and \
                    len(set([c[2] for c in cands])) == 1:
                # 品詞と品詞分類2が同じ場合、「一般」でないほう
                newcands = [c for c in cands if c[1] != "一般"]
                if newcands:
                    # 二種類しかないことをあてにしてる。よろしくないけど。
                    newattrs = newcands[0]
            elif len(set([c[0] for c in cands])) == 1:
                # 品詞が同じ場合、「一般」でないほう
                newcands = [c for c in cands if c[1] != "一般"]
                if newcands:
                    # 同上。
                    newattrs = newcands[0]
            elif len(set([c[0] for c in cands])) == 1 and \
                    len(set([c[1] for c in cands])) == 1:
                # 品詞と品詞分類1が同じ場合、「一般」でないほう
                newcands = [c for c in cands if c[2] != "一般"]
                if newcands:
                    # 同上。
                    newattrs = newcands[0]
            elif len(set([c[0] for c in cands])) > 1:
                # 品詞が違う場合…、うーん、名詞、かなぁ?
                # サ変接続 vs 副詞/副詞可能/助詞類接続/形容動詞語幹
                # だともう用例そのものが違うから、本来両方いていい
                # パターンなんだけれど。
                newcands = [c for c in cands if c[0] == "名詞"]
                if newcands:
                    # 同上。
                    newattrs = newcands[0]
            if not newattrs:
                logging.info("%r", list(cands))
        #
        class _hinshi_update(object):
            def __init__(self, na):
                self._na = na
            def __call__(self, ent):
                for j, k in enumerate(newattrs_keys):
                    ent[k] = self._na[j]
                return ent
        #
        if gbkk2[0][0][0] == "*" and newattrs:
            bfi["hinshi_update"] = _hinshi_update(newattrs)
        else:
            bfi["hinshi_update"] = lambda ent: ent
        #
        return bfi

    def _variation_kind(self, bfi, i, ent):
        # 合致等々についての情報収集に基いて、エントリを調整する。
        # 分類が主だが、品詞の読み替えも行なう。
        sur = ent["表層形"]
        of = ent["原形"]
        dh = ent["代表表記"]
        dhy = ent["代表表記(読み)"]
        ofy = ent["原形(読み)"]
        rank = ent["同一原形(読み)内順位"]

        # ---------------------------------------------
        #
        # まずは jumandic 内だけで閉じた部分。
        #
        res = []
        #
        ap0 = []
        dhy_n, dhy = dhy.split(" ")  # 「1/2 イイガタイ」など
        ofy_n, ofy = ofy.split(" ")  # 「2/2 スッゴイ」など
        ofun = _rm_alnum(of)
        kvsskey = ent["原形(読み)"].split(" ")[1]
        kvss = bfi["kvss"][kvsskey]
        if dhy_n != "1/1":
            ap0.append("読み" + dhy_n)
        if ofy_n != "1/1" and ofy_n != dhy_n:
            ap0.append("変種" + ofy_n)
        if ap0:
            res.append("(:{}:)".format("|".join(ap0)))
        if rank != "1/1":
            res.append("(:同原形(読み)内ランク{}:)".format(rank))
        cl1, cl2 = kvss.classified(of)
        if cl1:
            res.append("(:仮名語変種({}):)".format("|".join(cl1)))
        elif cl2:
            res.append("(:変種({}):)".format("|".join(cl2)))
        if kvss.has_dkanji() == "K":
            res.append("(:異漢字変種あり:)")
        elif kvss.has_dkanji() == "A":
            # 「平方キロメートル/ｋｍ３」パターン。
            res.append("(:かな漢字/英数変種あり:)")

        ent[":コメント:"] += "".join(res)

        # ---------------------------------------------
        #
        # 続いて、ipadic との合致に関係する部分。
        #
        ap = []
        if not bfi["allhira"] and bfi["matched_nonhira"]:
            bfi["hinshi_update"](ent)
            if bfi["vars"][i]["yv"]:
                dv = _yomi_diffvar(
                    ent["原形"],
                    ent["原形(読み)"].split(" ")[1],
                    bfi["vars"][i]["ma"][-1])
                if dv:
                    # これに該当するものは、わかりきっている「いらない」
                    # 変種と考えていい。ipadic のミスっぽいのもあるが、
                    # どうでもいい人にとってはどうでもいい。量は多くない
                    # ので、どうしても有用だと思う人は、個別に引っこ抜けば
                    # よろしい。
                    ap.append("読み[{}]".format(dv))
                elif "送り仮名" in ent[":コメント:"]:
                    # これに該当するものは、はっきりいって「迷惑」な
                    # エントリである。「垂れる」は「タレル」であろう、
                    # 「シダレル」ではない。(国語辞書に載ってる載ってない
                    # とは無関係な議論であることに注意。MeCab を実用的に
                    # 使う、というただ一点においては、「垂れる」に対する
                    # 「タレル/シダレル」という二重化は何の意味もない、
                    # という話。)
                    # 「読み[送り仮名]」という表現、わかりにくいかもね。
                    # 冗長な言い方をするならば、「送り仮名違いにより
                    # 偶然合致したものの ipadic が想定してるものとは別物
                    # だよヴぁーかヴぁーか」てことね。これも数は多くない
                    # ので、どういうことになってるのか、ご自分の目で全部
                    # 確かめてみたらよろしい。まったくの別語と合致してる
                    # ことがある。
                    ap.append("読み[送り仮名]")
                elif "仮名化" in ent[":コメント:"]:
                    # 該当は3つなのだが、「好き心」以外はよくわからん。
                    # ワタシ個人にとってははっきりいってどれも迷惑な
                    # エントリだが、おぬしはみーつーか? これもご自身で
                    # 判断されるが良かろう。
                    ap.append("読み[衝突?<代表表記:{}|ipadic:{}>]".format(
                            ent["代表表記"], bfi["vars"][i]["ma"][-1]))
                else:
                    ap.append("読み<{}>".format(
                            bfi["vars"][i]["ma"][-1]))
            if bfi["vars"][i]["vv"]:
                ap.append("活用")
            ap = ("({})".format("|".join(ap))) if ap else ""
            ent[":コメント:"] += "(:ipadic既存エントリ変種{}:)".format(ap)

    def _process_each_dh(self, gbkk2):
        bfi = self._basicform_info(gbkk2)
        for kk2, ents in gbkk2:
            for i, ent in enumerate(ents):
                if not ent.get(":コメント:"):
                    ent[":コメント:"] = ""
                self._variation_kind(bfi, i, ent)
                if bfi["vars"][i]["ac"]:
                    yield ent

    def __call__(self):
        # 合致・衝突有無で jumandic エントリを許容するかどうかという
        # ことだが、話はとても複雑。全体説明参照。
        # ・活用する語は、「基本形」をみる。
        #   活用の一部だけがマッチしないのは気にしないことにする。
        #   こういうのはつまり ipadic 流儀と異なる活用ということだから。
        #   活用の一部だけ取り込んでもそんなに価値はないだろう。
        # ・パターンは
        #   1. ipadic から見つからない、めでてーな
        #   2. ipadic から、バリエーションの一つだけが合致する
        #   3. ipadic から、バリエーションの複数が別々に合致する
        #      (さらに合致の各々が同じ場合と違う場合(合致の衝突))
        #   なお、「原形 == 代表形」を優先することは出来ない。必ずしも
        #   もっともメジャーな綴りが代表になってるとは限らない。
        #   (むしろヘンな綴りが選ばれてる傾向が高い。たぶん含まれる
        #   漢字の量で機械的に決めてるんだと思う。)
        # なお、「合致に基く品詞の読み替え」に関して、活用語は複雑
        # 過ぎるので、活用しない語の場合のみとする。(とりわけ非自立
        # と接尾に関係する読み替えはかなりシビア。)
        for dh, entries in itertools.groupby(
            self._reader, lambda x: (
                x["代表表記"],
                )):
            # 内側では、活用形2でまとめる。プリプロセスのソートで
            # 基本形が先頭になってる。
            gbkk2 = [
                (g[0], list(g[1]))
                for g in itertools.groupby(
                    entries, lambda e: (
                        e["活用形2"], e["-活用形2内部用-"]))]
            yield from self._process_each_dh(gbkk2)


# ==================================================================
# 第四段階関連のもろもろ
# ==================================================================
#
# 最終フェーズによる分類(書き出し単位)に関して。コード本体部分に
# 書くと相当長くて鬱陶しいのでここで。
#
# 分類の基準は原則としてたった二つ。「こんなんいるかーっ」となる
# 可能性が高いかどうかが一つ。そしてもちろん「問題を起こしやすいか
# どうか」。両者はもちろん根は同じだが、傷の深さは同じではない。
#
# 一つ目の基準の、異論はありそうだがわかりやすいのは「ipadic と
# 読みしか違わない変種」である。「読み」や「意味」が大事な何かを
# MeCab で行ないたいならこの変種は有用となりうる。けれども
# トークン分解・分かち書きだけを目的とする MeCab 使いにとっては
# この違いは無駄でしかなく、自己満足にしか感じない。要するに
# テリー並に「いらない子」である。けれどもこれは決して「有害」
# というものではなく、傷は極めて浅い。いらない人にとっては
# 「新婦るに、煎ら亡い」というだけのことである。
#
# 「固有名詞」というくくりそのものはあまり異論はないと信じてる。
# 色んな場所で書いてるとおり「辞書は自分の扱いたい問題領域に
# ふさわしいものを使うのがベスト」というのは自明の理であって、
# 固有名詞をがんがん放り込んだほうがいい場合とそうでない場合は
# 色々であるわけだし。あとね、jumandic の「組織」の考え方に
# 私は「システム辞書の内容としては」違和感がある。現実のテキスト
# では「駒沢」を大学と同義とみなすことが多いのだとしても、それは
# 「そういうドメインのテキスト」を扱いたい人がそうすればいいこと
# なだけで、システム辞書に入っていなければならない理由はないんじゃ
# ないのかなと。略称もかなり微妙。確かに「経済企画庁」を「経企」
# とするのは頻出なのかもしれない。けれども、、、以下同文である。
# システム辞書でやることではないんでは。(無論これがユーザ辞書
# としての追加エントリ候補として抽出されることそのものは尊い。)
#
# なんつーのかなぁ、ipadic も jumandic もそうなんだけれど、
# 「オプショナル」な考え方を持っててくれればね、もっと使いやすい
# 辞書になったんじゃないかと思うよ、ほんと。全部いっしょくたに
# しようとしてしまうことそのものが、「システム辞書」に馴染まない
# わけだよ。
#
# 一方もう一つの基準については、なかなか一筋縄ではいかないのだが、
# 「混同されるとイラっとする確率が非常に高い一群」はある。
# どのようなケースでもコストの調整を伴わずにこれの本質的な
# 解決は出来ず、すなわち「使うか使わないかの二択だが使うと
# 決めると地獄に落ちる」のでは困るので、コスト調整は不可避、
# だとしても、である。たとえば「保母」のひらがなバージョン。
# 何が起こるかわかるよな? こういうの、コスト調整して出来るだけ
# 選択されないようにする…以前に、「こんなエントリが入ってる
# からやろ」って思わん? 思うでしょう、思わないとは言わせない。
# すなわち、「固有名詞と、名詞-一般のひらがなバージョン」に
# よる被害は、かなりカチンと来るのである。コード内でこまごま説明
# するとうるさいので黙って処理しているが、いずれもこの観点で
# 「追い出している」(具体的には別フォルダに押し込める)。当然
# こうしたものは「ドメイン固有」として必要となることはあるわけだ。
# たとえば小学生以下の幼児の文章を解析・分析したいなら、こうした
# ひらがなエントリはあるは役に立つ可能性はある。要するに「やり
# たいこと次第」ってわけである。
#
# 非自立、接尾、接頭詞も追い出しているのだが、これはあると問答
# 無用で便利だと想像するかもしれない。けれどもこれは結構諸刃の
# 剣。やってみればわかるけれど、接尾は強過ぎて、色んなところで
# くっつこうとし過ぎる。短い語が多いから、なんだけどね、「-み」
# だのがあらゆるものにくっつきたがる。なので、実際解析させてみて、
# よろしくないようなら使わない方がいい。これはひらがなだとか
# 云々関係なく。(似たようなので、助詞、助動詞が微妙だ…。)
#

def _extract(fn, fromsysdic):
    results = defaultdict(list)
    ckeys = ("品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3")
    for line in _accumulator(fn, fromsysdic)():
        kind = []
        comment = line[":コメント:"]
        # 第一階層は ipadic に既存かどうか。
        kind.append("vars_of_ipadict" if "既存エントリ変種" in comment else "new_entries")
        # 既存の場合、読み違い(普通は意味も違う)による変種かそうでないか。
        if re.search(r"既存エントリ変種\([^()]*読み[^()]*\)", comment):
            kind.append("yomi")
        #
        if line["品詞"] == "記号":
            if "顔文字" in line["出典等"]:
                line[":コメント:"] = "(:顔文字:)"
                kind.append("symbol_emoticon")
            else:
                kind.append("symbol")
        else:
            m = re.search(r"変種\(([^()]+)\)", comment)
            if m:
                var = m.group(1)
                # 送り仮名違いも仮名化も、9割以上は「迷惑に感じる」エントリで
                # あることは間違いないけれど、問題を起こしやすいかどうかという
                # 意味だと、実際は感じるほどではない。実際は現実の適用対象に
                # 対して、ごくごく少数のエントリが「大規模に」被害をもたらす
                # だけであることが多い。典型的なのは接尾みたいな短いワードが
                # 「ありとあらゆるところで使われる」みたいなこと。なので、
                # 自身で理解して使う分には見かけほどは大問題ではない…、ので、
                # まぁ要するに「大量にあるけど少しは分類したので、もしここから
                # 欲しいもんがあるよーなら自分で発掘せぃ」ちぅことね。ワタシは
                # いらんのだけどね。けど、どっかでも書いた通り、「漢字化してある
                # 方がへんちくりん」パターンは結構あるのでね、ワタシのこの決断が
                # 邪魔くさい人はぜってーいるハズ。
                if ("送り仮名" in var and "仮名化" not in var):
                    kind.append("contraversy1")
                elif ("送り仮名" not in var and "仮名化" in var):
                    kind.append("contraversy2")
                elif ("送り仮名" in var and "仮名化" in var):
                    kind.append("contraversy3")
            if line["品詞"] == "名詞" and (line["品詞細分類1"] in ("固有名詞", "一般")):
                if line["品詞細分類1"] == "固有名詞":
                    ki = "noun_proper"
                else:
                    ki = "noun_general"
                if not re.search(r"[^ぁ-んー・]", line["表層形"]):
                    ki += "_hiragana"
                kind.append(ki)
            #
            elif line["品詞細分類1"] in ("非自立", "接尾",) or line["品詞"] == "接頭詞":
                kind.append("depends")
        #
        kind = tuple(kind)

        for mvals in sorted(_omap.keys(), key=lambda x: (-len(x), x)):
            omkbase = tuple((line[k] for i, k in enumerate(ckeys) if i < len(mvals)))
            if omkbase == mvals:
                results[omkbase + (kind,)].append(line)
                break
    return results


def _dump(args, results):
    for ck in results.keys():
        entries = results[ck]
        if ck[:-1] not in _omap:
            continue
        ofnbase, keys = _omap[ck[:-1]]
        kind = ck[-1]
        ofn = "{}.from_jumandic.csv".format(ofnbase)
        dn = posixpath.join(*((args.distdir,) + kind))
        if not os.path.exists(dn):
            os.makedirs(dn)
        ofn = posixpath.join(dn, ofn)
        # jumandic 辞書エントリが複数ファイルにまたがっている
        # ことと、このスクリプト自身による再構成によって、
        # どうしてもこの時点で重複が残っている。ので、改めて
        # 重複を取り除きつつソートし直す。
        # (「:コメント:」についても、別々のルートから個々に
        # 違う主張をしてしまうことがあるので、これもまとめ
        # 直す。ひらがなオンリーのエントリでこういうことが
        # 起こる。それだけ色んな衝突をしてしまうということ。)
        ukeys = (keys + ("原形", ":コメント:",))
        unir = set([tuple((ent[k] for k in ukeys))
                    for ent in entries])
        pkeys = (
            "品詞", "品詞細分類1", "品詞細分類2", "品詞細分類3",
            "活用形1", "原形", "読み", "活用形2", "表層形")
        cleaned = []
        for g in itertools.groupby(
            sorted(
                list(unir),
                key=lambda x: tuple((x[ukeys.index(k)] for k in pkeys if k in ukeys))),
            lambda x: x[:-1]):  # 「:コメント:」列以外でグルーピング。
            #
            entrycols = list(g[0])  # 「:コメント:」列以外
            # 「:コメント:」列を一つにまとめる。
            comments = set()
            for e in list(g[1]):
                comments |= set(re.findall(r"\(:.*?:\)", e[-1]))
            cleaned.append(entrycols + ["".join(sorted(comments))])

        logging.info("writing result %r", ofn)
        with io.open(ofn, "wb") as fo:
            fo.write((", ".join(ukeys) + "\n").encode("utf-8"))
            fo.write(("\n".join([",".join(e) for e in cleaned]) + "\n").encode("utf-8"))


# ==================================================================
# コントローラ
# ==================================================================
def _fix_ipadic_entry(ipaent):
    # ipadic は割と手作業感満載なミスが多い。読みのミスばかり
    # だけれど。jumandic がやたらに「機械的」なのと対照的というか。
    # だからこそ「人の目が入ってる」安心感、なのだとは思う。
    # ともあれ、このワタシの「スクリプトだとかなんとかそいうの」
    # の処理の邪魔になるミスについては、補正しておく。無論処理に
    # 関係することしかやらない。ipadic を直す意図は毛頭ないので。

    # ipadic のミスのひとつに、本来全ての「読み」はカタカナ
    # として統一、というコンセプトらしいのに、なぜかひとつ
    # だけ読みにひらがなが混入している。この差異を jumandic
    # から取り込む…ことに到底意味があるとは思えないので、
    # これに関しては ipadic のミスをみなかったことにしておく。
    # (システム辞書メンテナンス用途にこの「スクリプトっぽい
    # それらしいもの」をあてにしている人のために一応言って
    # おくと、これは(jumandicとの合致から検出出来るものでは)
    # 人名の「かほり」のみ、かな、たぶん。)
    ipaent["読み"] = _h2k(ipaent["読み"])

    # 読みのミス。なお、ここで補正することにより、「jumandic
    # エントリによる修正」という使い方は出来なくなる。特に
    # 「不可逆」だが、ipadic は「フキャガク」という、誤って
    # いるけれども日常口頭語では良くやらかすやつを与えてて、
    # jumandic がまったくもって正しいのだけれど、でもさ、
    # 別にそれって「jumandic エントリを投入することによって」
    # やらんでもいいでしょ。
    # 
    if ipaent["読み"] == "ムトントン" and ipaent["発音"] == "ムトンチャク":
        ipaent["読み"] = "ムトンチャク"
    if ipaent["表層形"] == "不可逆":
        ipaent["読み"] = "フカギャク"
        ipaent["発音"] = "フカギャク"
    if ipaent["表層形"] == "人民元":
        ipaent["読み"] = "ジンミンゲン"
        ipaent["発音"] = "ジンミンゲン"
    if ipaent["表層形"] == "役どころ":
        ipaent["読み"] = "ヤクドコロ"
        ipaent["発音"] = "ヤクドコロ"
    if ipaent["表層形"].startswith("ぶっ殺"):
        ipaent["読み"] = ipaent["読み"].replace("ブットバ", "ブッコロ")
        ipaent["発音"] = ipaent["発音"].replace("ブットバ", "ブッコロ")
    return ipaent

def _main(args):
    global _yomi_norm

    fromsysdic_all = defaultdict(list)
    alphabets = {}
    for fn in glob(posixpath.join(_sysdict_dir, "*.csv")):
        reader = csv.DictReader(
            io.open(fn, encoding=_mecab_config["--sysdic-src-encoding"]),
            fieldnames=fieldnames_all)
        for line in reader:
            sur = line["表層形"]
            ent = _fix_ipadic_entry(line)
            fromsysdic_all[sur].append(ent)
            if ent["品詞"] == "記号" and \
                    ent["品詞細分類1"] == "アルファベット":
                alphabets[sur] = ent["読み"]
    _yomi_norm = __yomi_norm(alphabets)
    #
    _firststep(args)
    _secondstep(args)
    results = defaultdict(list)
    for fn in sorted(glob(posixpath.join(args.builddir, "2nd", "*.csv"))):
        r = _extract(fn.replace("\\", "/"), fromsysdic_all)
        for k in r.keys():
            results[k].extend(r[k])
    _dump(args, results)


# ==================================================================
# おメイン
# ==================================================================
if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--jumandicdir", default="mecab-jumandic-7.0-20130310")
    parser.add_argument("--distdir", default="_dist")
    parser.add_argument("--builddir", default="_build")
    args = parser.parse_args()
    _main(args)
